from napiod import model
import numpy as np
import copy
import bisect
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
from mpoints.plot_tools import (
    sample_path as hawkes_sample,
    kernels_exp as norms_of_hawkes_kernels,
)


def price_path(
        times,
        events,
        states,
        time_start,
        time_end,
        initial_price=0,
        num_subplots=2,
):
    # Reconstruct the price path from the states of the model
    prices = model.price_path_from_states(initial_price, states)
    # Reconstruct the eveolution of volume imbalances from the states of the model
    imbalances = model.imbalances_from_states(states)
    t0 = bisect.bisect_left(times, time_start)
    t1 = bisect.bisect_right(times, time_end)
    prices = list(copy.copy(prices[t0: t1]))
    imbalances = list(copy.copy(imbalances[t0: t1]))
    times = list(copy.copy(times[t0: t1]))

    f, axes = plt.subplots(num_subplots, 1, sharex='col')

    # Price path
    ax_price = axes[0]
    ax_price.step(times, prices, where='post', linewidth=2, label='price')
    ax_price.set_ylabel('Tick')
    ax_price.set_xlabel('Time')
    ax_price.legend()

    # Imabalance path
    ax_imb = axes[1]
    ax_imb.step(times, imbalances, where='post',
                linewidth=2, label='volume imbalance', color='darkgreen')
    ax_imb.set_xlabel('Time')
    ax_imb.set_ylabel('Volume imbalance')
    ax_imb.set_yticks([-1, 0, 1])
    ax_imb.set_yticklabels(['negative', 'neutral', 'positive'])

    plt.tight_layout()
    ax_imb.legend()
    return f, axes


def price_impact_profile(
        profile,
        time_start,
        time_end,
):
    t0 = bisect.bisect(profile[:, 0], time_start)
    t1 = bisect.bisect(profile[:, 0], time_end)

    f, axes = plt.subplots(3, 1, sharex='col')

    ax_tot = axes[0]
    ax_tot.plot(
        profile[t0:t1, 0],
        profile[t0:t1, 3],
        linewidth=2,
        color='blue',
        label='total impact'
    )
    ax_tot.legend()

    ax_direct = axes[1]
    ax_direct.plot(
        profile[t0:t1, 0],
        profile[t0:t1, 1],
        linewidth=2,
        color='darkorange',
        label='direct impact'
    )
    ax_direct.legend()

    ax_indirect = axes[2]
    ax_indirect.plot(
        profile[t0:t1, 0],
        profile[t0:t1, 2],
        linewidth=2,
        color='darkgreen',
        label='indirect impact'
    )
    ax_indirect.legend()
    ax_indirect.set_xlabel('Time')

    return f, axes


def agent_execution(
        times,
        events,
        time_start,
        time_end,
        fig=None,
):
    is_agent = events == model.AGENT_INDEX
    agent_times = np.array(
        times[is_agent],
        copy=True)
    agent_counter = 1 + np.arange(np.sum(is_agent))
    t0 = bisect.bisect(agent_times, time_start)
    t1 = bisect.bisect_right(agent_times, time_end)
    pret0 = max(0, t0 - 1)
    pret1 = max(pret0, t1 - 1)
    plot_times = np.concatenate(
        ([time_start], agent_times[t0: t1], [time_end]))
    plot_counter = np.concatenate(
        (agent_counter[[pret0]], agent_counter[t0: t1], agent_counter[[pret1]]))

    if fig is None:
        fig, axes = plt.subplots(1, 1)
        ax = axes
    else:
        ax = fig.axes[-1]

    ax.step(plot_times, plot_counter, where='post', color='red', label="agent")
    ax.set_xlabel("Time")
    ax.set_ylabel("Number of market orders")
    ax.legend()

    return fig, fig.axes


def price_path_and_agent_execution(
        times,
        events,
        states,
        time_start,
        time_end,
        initial_price=0,
):
    f, axes = price_path(
        times,
        events,
        states,
        time_start,
        time_end,
        initial_price,
        3
    )
    f, axes = agent_execution(
        times,
        events,
        time_start,
        time_end,
        fig=f
    )
    return f, axes



import setuptools
import numpy

ext_modules = [
    setuptools.Extension('mpoints.hybrid_hawkes_exp_cython',
                         sources=['mpoints/hybrid_hawkes_exp_cython.c'],
                         extra_compile_args=["-ffast-math"],
                         )
]

setuptools.setup(
    name="mpoints",
    ext_modules=ext_modules,
    include_dirs=[numpy.get_include()],
)



import numpy as np
import scipy
from mpoints import hybrid_hawkes_exp


def produce_samples(
):
    number_of_event_types: int = 5
    de = number_of_event_types
    number_of_states: int = 9
    dx = number_of_states
    events_labels = [chr(65 + n) for n in range(number_of_event_types)]
    states_labels = [chr(48 + n) for n in range(number_of_states)]
    _phis = [np.eye(dx) + scipy.sparse.random(dx, dx,
                                              density=.50).A for _ in range(de)]
    _phis = [_phi / 10*np.sum(_phi, axis=1, keepdims=True)
             for _phi in _phis]
    _phis = [np.expand_dims(_phi, axis=1) for _phi in _phis]
    phis = np.concatenate(_phis, axis=1)
    nus = np.random.uniform(low=0., high=1., size=(de,))
    alphas = .5 * scipy.sparse.random(de, dx*de, density=.50).A.reshape(de, dx, de)
    betas = np.ones((de, dx, de), dtype=float)
    hhe = hybrid_hawkes_exp.HybridHawkesExp(
        number_of_event_types,
        number_of_states,
        events_labels,
        states_labels,
    )
    hhe.set_transition_probabilities(phis)
    hhe.set_hawkes_parameters(nus, alphas, betas)
    time_start = 0.
    time_end = 10000.
    times, events, states = hhe.simulate(
        time_start, time_end, max_number_of_events=1000000)
    return hhe, times, events, states



import numpy as np
import napiod
from napiod.model import Constants
import matplotlib.pyplot as plt


AGENT_DIRECTION = -1  # The agents is selling


def balanced_transition_probabilities():
    phis = np.zeros(
        Constants.shape_of_transition_matrix(),
        dtype=float
    )

    # Event type 1: sell market order
    phis[np.ix_(
        Constants.negative_imbalance_states(),
        [1],
        Constants.deflationary_states())
    ] = 1. / len(Constants.deflationary_states())
    phis[np.ix_(
        Constants.non_negative_imbalance_states(),
        [1],
        Constants.non_inflationary_states())
    ] = 1. / len(Constants.non_inflationary_states())

    # Event type 2: buy market order
    phis[np.ix_(
        Constants.positive_imbalance_states(),
        [2],
        Constants.inflationary_states())
    ] = 1. / len(Constants.inflationary_states())
    phis[np.ix_(
        Constants.non_positive_imbalance_states(),
        [2],
        Constants.non_deflationary_states())
    ] = 1. / len(Constants.non_deflationary_states())

    # Event type 3
    phis[np.ix_(
        np.arange(phis.shape[0]),
        [3],
        Constants.deflationary_states())
    ] = 1. / len(Constants.deflationary_states())

    # Event type 4
    phis[np.ix_(
        np.arange(phis.shape[0]),
        [4],
        Constants.inflationary_states())
    ] = 1. / len(Constants.inflationary_states())

    return phis


def balanced_hawkes_coefficients():
    # base rates
    nus = .2 * np.ones(napiod.model.TOTAL_NUMBER_OF_EVENT_TYPES, dtype=float)

    # impact coefficients
    alphas = np.zeros(
        Constants.shape_of_impact_coefficients(),
        dtype=float
    )
    alphas[np.ix_(
        np.arange(alphas.shape[0]),
        Constants.positive_imbalance_states(),
        Constants.non_deflationary_events())] = 1.
    alphas[np.ix_(
        np.arange(alphas.shape[0]),
        Constants.negative_imbalance_states(),
        Constants.non_inflationary_events())] = 1.

    # decay coefficients
    betas = 4. * np.ones(
        Constants.shape_of_impact_coefficients(),
        dtype=float
    )
    betas[np.ix_(
        np.arange(alphas.shape[0]),
        Constants.positive_imbalance_states(),
        Constants.non_deflationary_events())] = 3.
    betas[np.ix_(
        np.arange(alphas.shape[0]),
        Constants.negative_imbalance_states(),
        Constants.non_inflationary_events())] = 3.

    return nus, alphas, betas


def bull_unbalanced_hawkes_coefficients():

    # base rates
    nus = np.zeros(napiod.model.TOTAL_NUMBER_OF_EVENT_TYPES, dtype=float)
    nus[Constants.non_inflationary_events()] = .20
    nus[Constants.non_deflationary_events()] = .25

    # impact coefficients
    alphas = np.zeros(
        Constants.shape_of_impact_coefficients(),
        dtype=float
    )
    alphas[np.ix_(
        np.arange(alphas.shape[0]),
        Constants.positive_imbalance_states(),
        Constants.non_deflationary_events())] = 1.5
    alphas[np.ix_(
        np.arange(alphas.shape[0]),
        Constants.negative_imbalance_states(),
        Constants.non_inflationary_events())] = 0.75

    # decay coefficients
    betas = 4. * np.ones(
        Constants.shape_of_impact_coefficients(),
        dtype=float
    )
    betas[np.ix_(
        np.arange(alphas.shape[0]),
        Constants.positive_imbalance_states(),
        Constants.non_deflationary_events())] = 2.95
    betas[np.ix_(
        np.arange(alphas.shape[0]),
        Constants.negative_imbalance_states(),
        Constants.non_inflationary_events())] = 3.25

    return nus, alphas, betas


def bear_unbalanced_hawkes_coefficients():

    # base rates
    nus = np.zeros(napiod.model.TOTAL_NUMBER_OF_EVENT_TYPES, dtype=float)
    nus[Constants.non_inflationary_events()] = .25
    nus[Constants.non_deflationary_events()] = .20

    # impact coefficients
    alphas = np.zeros(
        Constants.shape_of_impact_coefficients(),
        dtype=float
    )
    alphas[np.ix_(
        np.arange(alphas.shape[0]),
        Constants.positive_imbalance_states(),
        Constants.non_deflationary_events())] = 0.75
    alphas[np.ix_(
        np.arange(alphas.shape[0]),
        Constants.negative_imbalance_states(),
        Constants.non_inflationary_events())] = 1.5

    # decay coefficients
    betas = 4. * np.ones(
        Constants.shape_of_impact_coefficients(),
        dtype=float
    )
    betas[np.ix_(
        np.arange(alphas.shape[0]),
        Constants.positive_imbalance_states(),
        Constants.non_deflationary_events())] = 3.25
    betas[np.ix_(
        np.arange(alphas.shape[0]),
        Constants.negative_imbalance_states(),
        Constants.non_inflationary_events())] = 2.95

    return nus, alphas, betas


def balanced_order_book():
    # low volatility
    phis = balanced_transition_probabilities()
    nus, alphas, betas = balanced_hawkes_coefficients()
    return (
        phis[:, 1:, :],
        nus[1:],
        alphas[1:, :, 1:],
        betas[1:, :, 1:],
    )


def bull_unbalanced_order_book():
    # high volatility
    phis = balanced_transition_probabilities()
    nus, alphas, betas = bull_unbalanced_hawkes_coefficients()
    return (
        phis[:, 1:, :],
        nus[1:],
        alphas[1:, :, 1:],
        betas[1:, :, 1:],
    )


def bear_unbalanced_order_book():
    # high volatility
    phis = balanced_transition_probabilities()
    nus, alphas, betas = bear_unbalanced_hawkes_coefficients()
    return (
        phis[:, 1:, :],
        nus[1:],
        alphas[1:, :, 1:],
        betas[1:, :, 1:],
    )


def agent_walks_the_book():
    agent_state_transition = np.zeros(
        (napiod.model.TOTAL_NUMBER_OF_STATES,
         napiod.model.TOTAL_NUMBER_OF_STATES),
        dtype=float)
    agent_state_transition[np.ix_(
        Constants.negative_imbalance_states(),
        Constants.deflationary_states())
    ] = 1. / len(Constants.deflationary_states())
    agent_state_transition[np.ix_(
        Constants.non_negative_imbalance_states(),
        Constants.non_inflationary_states())
    ] = 1. / len(Constants.non_inflationary_states())
    return agent_state_transition


def trend_following_agent():
    base_rate = .2
    agent_self_excitation = np.zeros(
        Constants.shape_of_agent_self_excitation(),
        dtype=float)
    agent_self_excitation[np.ix_(
        Constants.deflationary_events(),
        Constants.negative_imbalance_states())
    ] = 2.
    agent_self_excitation[np.ix_(
        Constants.deflationary_events(),
        Constants.neutral_imbalance_states())
    ] = 0.5
    agent_self_decay = 4. * np.ones(
        Constants.shape_of_agent_self_decay(),
        dtype=float)
    return base_rate, agent_self_excitation, agent_self_decay


def contrarian_agent():
    base_rate = .2
    agent_self_excitation = np.zeros(
        Constants.shape_of_agent_self_excitation(),
        dtype=float)
    agent_self_excitation[np.ix_(
        Constants.non_deflationary_events(),
        Constants.positive_imbalance_states())
    ] = 2.
    agent_self_excitation[np.ix_(
        Constants.non_deflationary_events(),
        Constants.neutral_imbalance_states())
    ] = 1.
    agent_self_decay = 4. * np.ones(
        Constants.shape_of_agent_self_decay(),
        dtype=float)
    return base_rate, agent_self_excitation, agent_self_decay


def market_provides_liquidity():
    agent_impact_on_others = np.zeros(
        Constants.shape_of_agent_impact_on_others(),
        dtype=float)
    agent_impact_on_others[np.ix_(
        np.arange(agent_impact_on_others.shape[0]),
        -1 + Constants.non_deflationary_events())
    ] = 1.5
    agent_decay_on_others = 4.25 * np.ones(
        Constants.shape_of_agent_decay_on_others(),
        dtype=float)
    return agent_impact_on_others, agent_decay_on_others


def market_withdraws_liquidity():
    agent_impact_on_others = np.zeros(
        Constants.shape_of_agent_impact_on_others(),
        dtype=float)
    agent_impact_on_others[np.ix_(
        np.arange(agent_impact_on_others.shape[0]),
        -1 + Constants.non_inflationary_events())
    ] = 2.
    agent_decay_on_others = 4. * np.ones(
        Constants.shape_of_agent_decay_on_others(),
        dtype=float)
    return agent_impact_on_others, agent_decay_on_others


def run(
        order_book_params,
        agent_induced_state_transitions,
        agent_self,
        agent_on_others,
        title_of_run,
):

    # Get parameters of order book
    phis, nus, alphas, betas = order_book_params()

    # Get parameters of agent
    agent_state_transition = agent_induced_state_transitions()
    agent_base_rate, agent_self_excitation, agent_self_decay = agent_self()
    agent_impact_on_others, agent_decay_on_others = agent_on_others()

    # Instantiate PriceImpact and set all parameters
    price_impact = napiod.model.PriceImpact(AGENT_DIRECTION)
    price_impact.set_orderbook_and_agent(
        phis,
        nus,
        alphas,
        betas,
        agent_state_transition,
        agent_base_rate,
        agent_impact_on_others,
        agent_self_excitation,
        agent_decay_on_others,
        agent_self_decay,
    )

    # Simulate execution
    execution_start = 0.
    execution_end = 100.
    horizon = 200.
    initial_state = 4
    times, events, states = price_impact.simulate_execution(
        execution_start,
        execution_end,
        horizon,
        initial_state=initial_state
    )

    # Compute price impact profile
    profile = price_impact.compute_price_impact_profile(
        times,
        events,
        states,
        horizon,
        execution_end,
        dt=.1 * times[0]
    )

    # Visualisations
    plot_time_start = 0.
    plot_time_end = horizon

    fig_price, _ = napiod.plot.price_path_and_agent_execution(
        times,
        events,
        states,
        plot_time_start,
        plot_time_end
    )
    fig_price.suptitle(title_of_run)

    fig_impact, _ = napiod.plot.price_impact_profile(
        profile,
        plot_time_start,
        plot_time_end
    )
    fig_impact.suptitle(title_of_run)

    return fig_price, fig_impact


def main():

    _, _ = run(
        balanced_order_book,
        agent_walks_the_book,
        trend_following_agent,
        market_provides_liquidity,
        'Trend following seller in balanced liquid market'
    )

    _, _ = run(
        balanced_order_book,
        agent_walks_the_book,
        trend_following_agent,
        market_withdraws_liquidity,
        'Trend following seller in balanced illiquid market'
    )

    _, _ = run(
        bear_unbalanced_order_book,
        agent_walks_the_book,
        trend_following_agent,
        market_withdraws_liquidity,
        'Trend following seller in bear unbalanced illiquid market'
    )

    _, _ = run(
        bull_unbalanced_order_book,
        agent_walks_the_book,
        contrarian_agent,
        market_provides_liquidity,
        'Contrarian seller in bull unbalanced liquid market'
    )

    plt.show()


if __name__ == '__main__':
    main()



from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext
import numpy

# Run in the console (under Anaconda version): python setup.py build_ext --inplace

# setup(
#     ext_modules=cythonize("hybrid_hawkes_exp_likelihood.pyx"),
#     include_dirs=[numpy.get_include()]
# )

ext_modules = [
    Extension(name="hybrid_hawkes_exp_cython",
              sources=["hybrid_hawkes_exp_cython.pyx"],
              # comment this line when compiling on Windows
              libraries=["m"],
              extra_compile_args=["-ffast-math"]
              )
]

setup(
    name="hybrid_hawkes_exp_cython",
    cmdclass={"build_ext": build_ext},
    ext_modules=ext_modules,
    include_dirs=[numpy.get_include()])



import unittest
import numpy as np
import bisect
import scipy
from mpoints import hybrid_hawkes_exp
from napiod import impact


class ImpactTest(unittest.TestCase):
    def setUp(self):
        number_of_event_types: int = 5
        de = number_of_event_types
        number_of_states: int = 9
        dx = number_of_states
        events_labels = [chr(65 + n) for n in range(number_of_event_types)]
        states_labels = [chr(48 + n) for n in range(number_of_states)]
        _phis = [np.eye(dx) + scipy.sparse.random(dx, dx,
                                                  density=.75).A for _ in range(de)]
        _phis = [_phi / np.sum(_phi, axis=1, keepdims=True)
                 for _phi in _phis]
        _phis = [np.expand_dims(_phi, axis=1) for _phi in _phis]
        phis = np.concatenate(_phis, axis=1)
        nus = np.random.uniform(low=0., high=1., size=(de,))
        alphas = .5 * \
            scipy.sparse.random(de, dx*de, density=.50).A.reshape(de, dx, de)
        betas = np.random.uniform(low=.95, high=1.05, size=(de, dx, de))
        hhe = hybrid_hawkes_exp.HybridHawkesExp(
            number_of_event_types,
            number_of_states,
            events_labels,
            states_labels,
        )
        hhe.set_transition_probabilities(phis)
        hhe.set_hawkes_parameters(nus, alphas, betas)
        time_start = 0.
        time_end = 10000.
        times, events, states = hhe.simulate(
            time_start, time_end, max_number_of_events=1000000)
        self.hhe = hhe
        self.number_of_event_types = number_of_event_types
        self.number_of_states = number_of_states
        self.phis = phis
        self.nus = nus
        self.alphas = alphas
        self.betas = betas
        self.times = times
        self.events = events
        self.states = states

    def test_pip(self):
        direction = -1
        j = np.random.randint(low=len(self.times) // 2, high=len(self.times))
        horizon = self.times[j]
        execution_end = self.times[-1]
        dt = self.times[0] / 2.
        self._test_pip(
            direction,
            horizon,
            execution_end,
            dt,
        )

    def _test_pip(self,
                  direction,
                  horizon,
                  execution_end,
                  dt):

        pip = impact.price_impact_profile(
            direction,
            horizon,
            execution_end,
            dt,
            self.phis,
            self.nus,
            self.alphas,
            self.betas,
            self.times,
            self.events,
            self.states,
        )

        idx_half = len(pip) // 2
        half_horizon = pip[idx_half, 0]

        half = impact.price_impact_profile(
            direction,
            half_horizon,
            execution_end,
            dt,
            self.phis,
            self.nus,
            self.alphas,
            self.betas,
            self.times,
            self.events,
            self.states,
        )
#         print(f'pip.shape: {pip.shape}')
#         print(f'half.shape: {half.shape}')
        self.assertEqual(
            pip.shape,
            (len(pip), 4)
        )
        self.assertEqual(
            half.shape,
            (1 + idx_half, 4)
        )
        self.assertTrue(
            np.all(
                np.diff(pip[:, 0]) > 0.
            ),
            '\nTimes in the price impact profiles are not increasing!'
        )
        self.assertTrue(
            np.all(
                np.diff(pip[:, 1]) >= 0.
            ),
            '\nThe profile of the direct price impact is decreasing!'
        )
        self.assertTrue(
            np.allclose(
                pip[:, 3],
                pip[:, 1] + pip[:, 2],
                rtol=1e-6,
                atol=1e-12,
            ),
            '\nOverall impact is not the sum of direct and indirect impacts'
        )
        self.assertTrue(
            np.allclose(
                half,
                pip[:len(half), :],
                rtol=1e-6,
                atol=1e-12,
            )
        )
        t1 = bisect.bisect_right(pip[:, 0], execution_end)
        self.assertTrue(
            np.allclose(
                np.diff(pip[t1:, 1]),
                0.
            ),
            'Direct impact is not flat after execution'
        )


if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np
import scipy
import bisect

from mpoints import hybrid_hawkes_exp
from napiod import model


class ModelTest(unittest.TestCase):
    def setUp(self):
        number_of_event_types: int = model.NUMBER_OF_ORDERBOOK_EVENTS
        de = number_of_event_types
        number_of_states: int = model.TOTAL_NUMBER_OF_STATES
        dx = number_of_states
        events_labels = [
            'sell_mo',
            'buy_mo',
            'deflo',
            'inflo',
        ]
        states_labels = [
            'd_neg',
            'd_neu',
            'd_pos',
            'f_neg',
            'f_neu',
            'f_pos',
            'u_neg',
            'u_neu',
            'u_pos',
        ]
        phis = model.generate_random_transition_probabilities(with_agent=False)
        nus = np.random.uniform(low=0., high=1., size=(de,))
        alphas = .5 * \
            scipy.sparse.random(de, dx*de, density=.50).A.reshape(de, dx, de)
        betas = np.random.uniform(low=.95, high=1.05, size=(de, dx, de))
        # `hhe` represents the order book without the agent
        hhe = hybrid_hawkes_exp.HybridHawkesExp(
            number_of_event_types,
            number_of_states,
            events_labels,
            states_labels,
        )
        hhe.set_transition_probabilities(phis)
        hhe.set_hawkes_parameters(nus, alphas, betas)
        self.hhe = hhe

    def test_labels(self):
        self.assertEqual(
            len(model.EVENTS),
            model.TOTAL_NUMBER_OF_EVENT_TYPES,
        )
        self.assertEqual(
            len(model.STATES),
            model.TOTAL_NUMBER_OF_STATES,
        )

    def test_add_seller_agent(self):
        direction = -1
        self._test_add_agent_params(direction)

    def test_add_buyer_agent(self):
        direction = 1
        self._test_add_agent_params(direction)

    def _test_add_agent_params(self, direction):
        idem_event = model.idem_event(direction)

        def _test(agent_base_rate=None):
            # Expand the parameters of the Hawkes process `hhe` in order to include the agent.
            phis, nus, alphas, betas = model._add_agent_params(
                direction,
                self.hhe.transition_probabilities,
                self.hhe.base_rates,
                self.hhe.impact_coefficients,
                self.hhe.decay_coefficients,
                agent_base_rate=agent_base_rate,
            )
            self.assertEqual(
                phis.shape,
                (model.TOTAL_NUMBER_OF_STATES,
                 model.TOTAL_NUMBER_OF_EVENT_TYPES,
                 model.TOTAL_NUMBER_OF_STATES),
            )
            self.assertEqual(
                nus.shape,
                (model.TOTAL_NUMBER_OF_EVENT_TYPES,),
            )
            self.assertEqual(
                alphas.shape,
                (model.TOTAL_NUMBER_OF_EVENT_TYPES,
                 model.TOTAL_NUMBER_OF_STATES,
                 model.TOTAL_NUMBER_OF_EVENT_TYPES
                 ),
            )
            self.assertEqual(
                betas.shape,
                (model.TOTAL_NUMBER_OF_EVENT_TYPES,
                 model.TOTAL_NUMBER_OF_STATES,
                 model.TOTAL_NUMBER_OF_EVENT_TYPES
                 ),
            )
            expected_base_rate = 0. if agent_base_rate is None else agent_base_rate
            self.assertTrue(
                np.allclose(
                    nus[0],
                    expected_base_rate,
                )
            )
            self.assertTrue(
                np.allclose(
                    alphas[0, :, 1:],
                    self.hhe.impact_coefficients[idem_event, :, :]
                )
            )
            self.assertTrue(
                np.allclose(
                    alphas[:, :, 0],
                    0.,
                )
            )
            self.assertTrue(
                np.allclose(
                    betas[0, :, 1:],
                    self.hhe.decay_coefficients[idem_event, :, :]
                )
            )

        # Test 1: we do not pass agent parameters, so they are all set to zero
        _test(None)

        # Test 2: Poisson execution
        _test(agent_base_rate=1.)

    def test_simulation_of_buyer_execution(self):
        direction = 1
        execution_start = 0.
        execution_end = 5000.
        horizon = 10000.
        self._test_simulation_of_execution(
            direction,
            execution_start,
            execution_end,
            horizon
        )

    def test_simulation_of_seller_execution(self):
        direction = -1
        execution_start = 0.
        execution_end = 5000.
        horizon = 10000.
        self._test_simulation_of_execution(
            direction,
            execution_start,
            execution_end,
            horizon
        )

    def _test_simulation_of_execution(
            self,
            direction,
            execution_start,
            execution_end,
            horizon,
    ):
        idem_event = model.idem_event(direction)
        # Poisson execution
        execution_base_rate = self.hhe.base_rates[idem_event]
        price_impact = model.PriceImpact(direction)
        price_impact.set_orderbook_and_agent(
            self.hhe.transition_probabilities,
            self.hhe.base_rates,
            self.hhe.impact_coefficients,
            self.hhe.decay_coefficients,
            agent_base_rate=execution_base_rate,
        )
        times, events, states = price_impact.simulate_execution(
            execution_start,
            execution_end,
            horizon,
            initial_state=1 + model.NUMBER_OF_IMBALANCE_SEGMENTS
        )
        self.assertTrue(
            np.all(np.diff(times) > 0.),
            'Event times are not increasing!',
        )
        self.assertTrue(
            np.all(
                events < model.TOTAL_NUMBER_OF_EVENT_TYPES
            ),
            'Unrecognised events in simulation'
        )
        self.assertTrue(
            np.all(
                states < model.TOTAL_NUMBER_OF_STATES
            ),
            'Unrecognised states in simulation'
        )
        t0 = bisect.bisect_left(times, execution_start)
        t1 = max(0, min(-1 + len(times), bisect.bisect_left(times, execution_end)))
        self.assertTrue(
            np.all(
                events[:t0] != model.AGENT_INDEX
            ),
            'Agent events found before execution starts',
        )
        self.assertTrue(
            np.all(
                events[t1:] != model.AGENT_INDEX
            ),
            'Agent events found after execution ends',
        )

        # Check states after inflationary events
        inflationary_event = 4
        is_inflationary_event = events == inflationary_event
        expected_states = list(model.inflationary_state_indexes())
        self.assertTrue(
            set(states[is_inflationary_event]).issubset(set(expected_states)),
        )

        # Check states after deflationary events
        deflationary_event = 3
        is_deflationary_event = events == deflationary_event
        expected_states = list(model.deflationary_state_indexes())
        self.assertTrue(
            set(states[is_deflationary_event]).issubset(set(expected_states)),
        )

        # Check states after agent's executions
        is_agent_event = events == model.AGENT_INDEX
        expected_states = list(model.stationary_state_indexes())
        if direction > 0:
            expected_states += list(model.inflationary_state_indexes())
        else:
            expected_states += list(model.deflationary_state_indexes())
        self.assertTrue(
            set(states[is_agent_event]).issubset(set(expected_states)),
        )

    def test_price_path_with_buyer(self):
        direction = 1
        execution_start = 0.
        execution_end = 5000
        horizon = 10000
        self._test_price_path(
            direction,
            execution_start,
            execution_end,
            horizon
        )

    def test_price_path_with_seller(self):
        direction = -1
        execution_start = 0.
        execution_end = 5000
        horizon = 10000
        self._test_price_path(
            direction,
            execution_start,
            execution_end,
            horizon
        )

    def _test_price_path(
            self,
            direction,
            execution_start,
            execution_end,
            horizon,
    ):
        idem_event = model.idem_event(direction)
        # Poisson execution
        execution_base_rate = self.hhe.base_rates[idem_event]
        price_impact = model.PriceImpact(direction)
        price_impact.set_orderbook_and_agent(
            self.hhe.transition_probabilities,
            self.hhe.base_rates,
            self.hhe.impact_coefficients,
            self.hhe.decay_coefficients,
            agent_base_rate=execution_base_rate,
        )
        times, events, states = price_impact.simulate_execution(
            execution_start,
            execution_end,
            horizon,
            initial_state=1 + model.NUMBER_OF_IMBALANCE_SEGMENTS
        )

        price_deltas = model.price_changes_from_states(states)
        self.assertEqual(
            set(price_deltas),
            {-1, 0, +1},
            'Unexpected price deltas',
        )

        inflationary_events = [4]
        deflationary_events = [3]
        non_inflationary_events = [1, 3]
        non_deflationary_events = [2, 4]
        if direction < 0:
            non_inflationary_events += [0]
        else:
            non_deflationary_events += [0]

        for e in inflationary_events:
            is_inflationary_event = events == e
            self.assertTrue(
                np.all(
                    price_deltas[is_inflationary_event] == 1
                ),
                f'price does not increase when inflationary event {e} occurs'
            )

        for e in deflationary_events:
            is_deflationary_event = events == e
            self.assertTrue(
                np.all(
                    price_deltas[is_deflationary_event] == -1
                ),
                f'price does not decrease when deflationary event {e} occurs'
            )

        for e in non_deflationary_events:
            is_nondelf_event = events == e
            self.assertTrue(
                np.all(
                    price_deltas[is_nondelf_event] != -1
                ),
                f'Price decreases when non deflationary event {e} occurs'
            )

        for e in non_inflationary_events:
            is_noninfl_event = events == e
            self.assertTrue(
                np.all(
                    price_deltas[is_noninfl_event] != 1
                ),
                f'Price increases when non inflationary event {e} occurs'
            )

    def test_state_conversions(self):
        for d in [3, 5, 7, 9]:
            for p in [-1, 0, +1]:
                for i in range(- (d // 2), 1 + d//2):
                    x = model.state_variable_to_state_index(
                        p, i, d)
                    p_ = model.price_change_from_state_index(
                        x, d)
                    i_ = model.imbalance_from_state_index(x, d)
                    self.assertEqual(
                        p, p_)
                    self.assertEqual(i, i_)
        self.assertEqual(
            list(model.stationary_state_indexes()),
            list(model._stationary_state_indexes_recon()),
            'Failed to reconcile stationary states',
        )
        self.assertEqual(
            list(model.inflationary_state_indexes()),
            list(model._inflationary_state_indexes_recon()),
            'Failed to reconcile inflationary states',
        )
        self.assertEqual(
            list(model.deflationary_state_indexes()),
            list(model._deflationary_state_indexes_recon()),
            'Failed to reconcile deflationary states',
        )

    def test_direct_impact_of_seller(self):
        direction = -1
        execution_end = 5000.
        horizon = 10000.
        self._test_direct_impact(
            direction,
            execution_end,
            horizon,
        )

    def test_direct_impact_of_buyer(self):
        direction = 1
        execution_end = 5000.
        horizon = 10000.
        self._test_direct_impact(
            direction,
            execution_end,
            horizon,
        )

    def _test_direct_impact(
            self,
            direction,
            execution_end,
            horizon,
    ):
        execution_start = 0.
        idem_event = model.idem_event(direction)
        # Poisson execution
        execution_base_rate = self.hhe.base_rates[idem_event]
        price_impact = model.PriceImpact(direction)
        price_impact.set_orderbook_and_agent(
            self.hhe.transition_probabilities,
            self.hhe.base_rates,
            self.hhe.impact_coefficients,
            self.hhe.decay_coefficients,
            agent_base_rate=execution_base_rate,
        )
        times, events, states = price_impact.simulate_execution(
            execution_start,
            execution_end,
            horizon,
            initial_state=1 + model.NUMBER_OF_IMBALANCE_SEGMENTS
        )
        profile = price_impact.compute_price_impact_profile(
            times,
            events,
            states,
            horizon,
            execution_end
        )
        self.assertTrue(
            np.all(
                profile[:, 1] >= 0.,
            ),
            'Direct impact should be non-negative',
        )
        self.assertTrue(
            np.all(
                np.diff(profile[:, 1]) >= 0.,
            ),
            'Direct impact should be non-decreasing',
        )
        t1 = bisect.bisect_right(profile[:, 0], execution_end)
        self.assertTrue(
            np.allclose(
                np.diff(profile[t1:, 1]),
                0.
            ),
            'Direct impact is not flat after execution has finished'
        )

    def test_indirect_impact_of_seller(self):
        direction = -1
        execution_end = 5000.
        horizon = 10000.
        self._test_indirect_impact(
            direction,
            execution_end,
            horizon
        )

    def test_indirect_impact_of_buyer(self):
        direction = 1
        execution_end = 5000.
        horizon = 10000.
        self._test_indirect_impact(
            direction,
            execution_end,
            horizon
        )

    def _test_indirect_impact(
            self,
            direction,
            execution_end,
            horizon
    ):
        execution_start = 0.
        idem_event = model.idem_event(direction)
        # Test 1: No impact of agent on other orderbook events
        execution_base_rate = self.hhe.base_rates[idem_event]
        agent_impact_on_others = np.zeros(
            (model.TOTAL_NUMBER_OF_STATES,
             model.NUMBER_OF_ORDERBOOK_EVENTS
             ),
            dtype=float)
        price_impact = model.PriceImpact(direction)
        price_impact.set_orderbook_and_agent(
            self.hhe.transition_probabilities,
            self.hhe.base_rates,
            self.hhe.impact_coefficients,
            self.hhe.decay_coefficients,
            agent_base_rate=execution_base_rate,
            agent_impact_on_others=agent_impact_on_others,
        )
        times, events, states = price_impact.simulate_execution(
            execution_start,
            execution_end,
            horizon,
            initial_state=1 + model.NUMBER_OF_IMBALANCE_SEGMENTS
        )
        profile = price_impact.compute_price_impact_profile(
            times,
            events,
            states,
            horizon,
            execution_end
        )
        self.assertTrue(
            np.allclose(
                profile[:, 2],
                0.,
                rtol=1e-7,
                atol=1e-12,
            ),
            'Indirect impact should be null when the agent has no impact on other orderbook events'
        )

        # Test 2: Only impact on events in the same direction
        same_events = [0, 2] if direction < 0 else [1, 3]
        agent_impact_on_others = np.zeros(
            (model.TOTAL_NUMBER_OF_STATES,
             model.NUMBER_OF_ORDERBOOK_EVENTS
             ),
            dtype=float)
        agent_impact_on_others[:, same_events] = np.random.uniform(
            low=0.,
            high=2.,
            size=(model.TOTAL_NUMBER_OF_STATES, 2)
        )
        price_impact = model.PriceImpact(direction)
        price_impact.set_orderbook_and_agent(
            self.hhe.transition_probabilities,
            self.hhe.base_rates,
            self.hhe.impact_coefficients,
            self.hhe.decay_coefficients,
            agent_base_rate=execution_base_rate,
            agent_impact_on_others=agent_impact_on_others,
        )
        times, events, states = price_impact.simulate_execution(
            execution_start,
            execution_end,
            horizon,
            initial_state=1 + model.NUMBER_OF_IMBALANCE_SEGMENTS
        )
        profile = price_impact.compute_price_impact_profile(
            times,
            events,
            states,
            horizon,
            execution_end
        )
        self.assertTrue(
            np.all(
                profile[:, 2] >= 0.,
            ),
            'Indirect price impact should be non-negative when agent has impact only on same events',
        )

        # Test 3: Only impact on events in the opposite direction
        opposite_events = [0, 2] if direction > 0 else [1, 3]
        agent_impact_on_others = np.zeros(
            (model.TOTAL_NUMBER_OF_STATES,
             model.NUMBER_OF_ORDERBOOK_EVENTS
             ),
            dtype=float)
        agent_impact_on_others[:, opposite_events] = np.random.uniform(
            low=0.,
            high=2.,
            size=(model.TOTAL_NUMBER_OF_STATES, 2)
        )
        price_impact = model.PriceImpact(direction)
        price_impact.set_orderbook_and_agent(
            self.hhe.transition_probabilities,
            self.hhe.base_rates,
            self.hhe.impact_coefficients,
            self.hhe.decay_coefficients,
            agent_base_rate=execution_base_rate,
            agent_impact_on_others=agent_impact_on_others,
        )
        times, events, states = price_impact.simulate_execution(
            execution_start,
            execution_end,
            horizon,
            initial_state=1 + model.NUMBER_OF_IMBALANCE_SEGMENTS
        )
        profile = price_impact.compute_price_impact_profile(
            times,
            events,
            states,
            horizon,
            execution_end
        )
        self.assertTrue(
            np.all(
                profile[:, 2] <= 0.,
            ),
            'Indirect price impact should be non-positive when agent has  impact only on opposite events',
        )


if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np
import scipy
from mpoints import hybrid_hawkes_exp


class HybridHawkesExpTest(unittest.TestCase):
    def setUp(self):
        number_of_event_types: int = np.random.randint(low=1, high=3)
        de = number_of_event_types
        number_of_states: int = np.random.randint(low=1, high=3)
        dx = number_of_states
        events_labels = [chr(65 + n) for n in range(number_of_event_types)]
        states_labels = [chr(48 + n) for n in range(number_of_states)]
        _phis = [np.eye(dx) + scipy.sparse.random(dx, dx,
                                                  density=.50).A for _ in range(de)]
        _phis = [_phi / 10*np.sum(_phi, axis=1, keepdims=True)
                 for _phi in _phis]
        _phis = [np.expand_dims(_phi, axis=1) for _phi in _phis]
        phis = np.concatenate(_phis, axis=1)
        nus = np.random.uniform(low=0., high=1., size=(de,))
        alphas = 10 * \
            scipy.sparse.random(de, dx*de, density=.50).A.reshape(de, dx, de)
        betas = np.ones((de, dx, de), dtype=float)
        self.number_of_event_types = number_of_event_types
        self.number_of_states = number_of_states
        self.events_labels = events_labels
        self.states_labels = states_labels
        self.nus = nus
        self.alphas = alphas
        self.betas = betas
        self.phis = phis

    def test_init_and_setters(self):
        hhe = hybrid_hawkes_exp.HybridHawkesExp(
            self.number_of_event_types,
            self.number_of_states,
            self.events_labels,
            self.states_labels,
        )
        self.assertEqual(self.number_of_event_types, hhe.number_of_event_types)
        self.assertEqual(self.number_of_states, hhe.number_of_states)
        self.assertEqual(self.phis.shape, hhe.transition_probabilities.shape)
        self.assertEqual(self.nus.shape, hhe.base_rates.shape)
        self.assertEqual(self.alphas.shape, hhe.impact_coefficients.shape)
        self.assertEqual(self.betas.shape, hhe.decay_coefficients.shape)

        hhe.set_transition_probabilities(self.phis)
        hhe.set_hawkes_parameters(self.nus, self.alphas, self.betas)
        self.assertTrue(np.allclose(self.phis, hhe.transition_probabilities))
        self.assertTrue(np.allclose(self.nus, hhe.base_rates))
        self.assertTrue(np.allclose(self.alphas, hhe.impact_coefficients))
        self.assertTrue(np.allclose(self.betas, hhe.decay_coefficients))

    def test_flatten_parameters(self):
        flat_params = hybrid_hawkes_exp.HybridHawkesExp.parameters_to_array(
            self.nus,
            self.alphas,
            self.betas
        )
        nus, alphas, betas = hybrid_hawkes_exp.HybridHawkesExp.array_to_parameters(
            flat_params,
            self.number_of_event_types,
            self.number_of_states,
        )
        self.assertTrue(
            np.allclose(nus, self.nus)
        )
        self.assertTrue(
            np.allclose(alphas, self.alphas)
        )
        self.assertTrue(
            np.allclose(betas, self.betas)
        )

    def _instantiate_hhe(self):
        hhe = hybrid_hawkes_exp.HybridHawkesExp(
            self.number_of_event_types,
            self.number_of_states,
            self.events_labels,
            self.states_labels,
        )
        return hhe

    def test_simulate(self):
        # Simulate
        hhe = self._instantiate_hhe()
        hhe.set_transition_probabilities(self.phis)
        hhe.set_hawkes_parameters(self.nus, self.alphas, self.betas)
        time_start = 0.
        time_end = 7200.
        times, events, states = hhe.simulate(
            time_start, time_end, max_number_of_events=1000000)
        self.assertTrue(
            np.all(states < self.number_of_states)
        )
        self.assertTrue(
            np.all(events < self.number_of_event_types)
        )
        self.assertTrue(np.all(np.diff(times) > 0.))

    @unittest.skip
    def test_simulate_and_estimate(self):
        # Simulate
        hhe = self._instantiate_hhe()
        hhe.set_transition_probabilities(self.phis)
        hhe.set_hawkes_parameters(self.nus, self.alphas, self.betas)
        time_start = 0.
        time_end = 7200.
        times, events, states = hhe.simulate(
            time_start, time_end, max_number_of_events=1000000)
        self.assertTrue(
            np.all(states < self.number_of_states)
        )
        self.assertTrue(
            np.all(events < self.number_of_event_types)
        )
        self.assertTrue(np.all(np.diff(times) > 0.))

        # Estimate
        estimator = self._instantiate_hhe()
        guess = estimator.parameters_to_array(
            self.nus, self.alphas, self.betas)
        guess *= np.random.uniform(low=.85, high=1.15, size=guess.shape)
        guess += np.random.uniform(low=0,  high=.001, size=guess.shape)
        opt_result, initial_guess, initial_guess_kind = estimator.estimate_hawkes_parameters(
            times, events, states, time_start, time_end,
            given_guesses=[guess],
        )
        mle_estimate = opt_result.x
        nus, alphas, betas = estimator.array_to_parameters(
            mle_estimate, estimator.number_of_event_types, estimator.number_of_states)
        print(f'Estimated nus:\n{nus}')
        print(f'True nus:\n{self.nus}')
        print(f'Estimated alphas:\n{alphas}')
        print(f'True alphas:\n{self.alphas}')
        print(f'Estimated betas:\n{betas}')
        print(f'True betas:\n{self.betas}')
        self.assertTrue(
            np.allclose(
                nus, self.nus, atol=1e-5, rtol=1e-4)
        )
        self.assertTrue(
            np.allclose(
                alphas, self.alphas, atol=1e-5, rtol=1e-4)
        )
        self.assertTrue(
            np.allclose(
                betas, self.betas, atol=1e-5, rtol=1e-4)
        )


if __name__ == '__main__':
    unittest.main()



import unittest
import bisect
import numpy as np
import scipy
from mpoints import hybrid_hawkes_exp
from napiod import intensities


class IntensitiesTest(unittest.TestCase):
    def setUp(self):
        number_of_event_types: int = 4
        de = number_of_event_types
        number_of_states: int = 9
        dx = number_of_states
        events_labels = [chr(65 + n) for n in range(number_of_event_types)]
        states_labels = [chr(48 + n) for n in range(number_of_states)]
        _phis = [np.eye(dx) + scipy.sparse.random(dx, dx,
                                                  density=.75).A for _ in range(de)]
        _phis = [_phi / np.sum(_phi, axis=1, keepdims=True)
                 for _phi in _phis]
        _phis = [np.expand_dims(_phi, axis=1) for _phi in _phis]
        phis = np.concatenate(_phis, axis=1)
        nus = np.random.uniform(low=0., high=1., size=(de,))
        alphas = .5 * \
            scipy.sparse.random(de, dx*de, density=.50).A.reshape(de, dx, de)
        betas = np.random.uniform(low=.95, high=1.05, size=(de, dx, de))
        hhe = hybrid_hawkes_exp.HybridHawkesExp(
            number_of_event_types,
            number_of_states,
            events_labels,
            states_labels,
        )
        hhe.set_transition_probabilities(phis)
        hhe.set_hawkes_parameters(nus, alphas, betas)
        time_start = 0.
        time_end = 10000.
        times, events, states = hhe.simulate(
            time_start, time_end, max_number_of_events=1000000)
        self.hhe = hhe
        self.number_of_event_types = number_of_event_types
        self.number_of_states = number_of_states
        self.phis = phis
        self.nus = nus
        self.alphas = alphas
        self.betas = betas
        self.times = times
        self.events = events
        self.states = states

    def test_accumulator_calc(self):
        e: int = np.random.randint(low=0, high=self.number_of_event_types)
        x: int = np.random.randint(low=0, high=self.number_of_states)
        t0: float = .0
        t1: float = self.times[-1]
        acc = np.zeros_like(self.alphas)
        acc = intensities._accumulator(
            e,
            x,
            t0,
            t1,
            self.betas,
            self.times,
            self.events,
            self.states,
            acc,
        )
        # Test that accumulator returns an array of the correct shape
        expected_shape = (
            self.hhe.number_of_event_types,
            self.hhe.number_of_states,
            self.hhe.number_of_event_types
        )
        self.assertEqual(
            acc.shape,
            expected_shape
        )

        # Test that the intensities computed using the accumulator are decreasing between arrival times
        j = len(self.times) // 2
        t_j = self.times[j]
        t_j_next = self.times[j+1]
        tau = self.times[-1]
        s = (t_j + t_j_next) / 2.
        t = (s + t_j_next) / 2.
        lambdas_s = intensities._intensities_for_recon(
            s,
            self.nus,
            self.alphas,
            self.betas,
            self.times,
            self.events,
            self.states
        )
        lambdas_t = intensities._intensities_for_recon(
            t,
            self.nus,
            self.alphas,
            self.betas,
            self.times,
            self.events,
            self.states
        )
        self.assertTrue(
            np.all(lambdas_t < lambdas_s),
            'Intensities are not decreasing between arrival times'
        )

        # Test consistency with mpoints
        napiod_lambdas = intensities._intensities_for_recon(
            t1,
            self.nus,
            self.alphas,
            self.betas,
            self.times,
            self.events,
            self.states
        )
        _, mpoints_lambdas = self.hhe.intensities_of_events_at_times(
            np.array([t1]),
            self.times,
            self.events,
            self.states)
        mpoints_lambdas = np.squeeze(mpoints_lambdas)
#         # print(f'NAPIOD intensities:\n{napiod_lambdas}')
#         # print(f'MPOINTS intensities:\n{mpoints_lambdas}')
        self.assertTrue(
            np.allclose(
                napiod_lambdas,
                mpoints_lambdas,
                rtol=1e-4,
                atol=1e-8,
            )
        )

    def test_accumulator_step_at_start(self):
        e: int = self.events[0]
        x: int = self.states[0]
        t0: float = self.times[0] / 2.
        t1: float = (self.times[0] + self.times[1]) / 2.
        self._test_accumulator_step(e, x, t0, t1)
        e = self.number_of_event_types - 1 if e == 0 else e - 1
        x = self.number_of_states - 1 if x == 0 else x - 1
        self._test_accumulator_step(e, x, t0, t1)

    def test_accumulator_step_at_second_start(self):
        e: int = self.events[0]
        x: int = self.states[0]
        t0: float = (self.times[0] + self.times[1]) / 2.
        t1: float = (self.times[1] + self.times[2]) / 2.
        self._test_accumulator_step(e, x, t0, t1)
        e = self.number_of_event_types - 1 if e == 0 else e - 1
        x = self.number_of_states - 1 if x == 0 else x - 1
        self._test_accumulator_step(e, x, t0, t1)

    def test_accumulator_step_random(self):
        e: int = np.random.randint(low=0, high=self.number_of_event_types)
        x: int = np.random.randint(low=0, high=self.number_of_states)
        t1: float = 1e-9 + np.random.choice(self.times[len(self.times) // 2:])
        t0: float = t1 / 2.
        self._test_accumulator_step(e, x, t0, t1)

    def _test_accumulator_step(self, e, x,  t0, t1):

        def recon(s, t, acc_vector, acc_step):
            acc_vector = intensities._accumulator(
                e,
                x,
                s,
                t,
                self.betas,
                self.times,
                self.events,
                self.states,
                acc_vector
            )
            acc_step = intensities._acc_step_recon(
                e,
                x,
                s,
                t,
                self.betas,
                self.times,
                self.events,
                self.states,
                acc_step
            )
#             print(f'acc_vector:\n{acc_vector}')
#             print(f'acc_step:\n{acc_step}')
            error_msg = f'\nacc_vector and acc_step do not reconcile in the increment from s={s} to  t={t}\n'
            error_msg += f'acc_vector:\n{acc_vector}\n\n'
            error_msg += f'acc_step:\n{acc_step}\n\n'
            self.assertTrue(
                np.allclose(
                    acc_vector,
                    acc_step,
                    atol=1e-12,
                    rtol=1e-7,
                ),
                error_msg,
            )
            return acc_vector, acc_step
#         print('\n...Testing accumulator step.....')
#         print(f'e: {e}')
#         print(f'x: {x}')
#         print(f't0: {t0}')
#         print(f't1: {t1}')
        acc_vector = np.zeros_like(self.alphas)
        acc_step = np.zeros_like(self.alphas)
        acc_vector, acc_step = recon(
            0., t0, acc_vector, acc_step)
        acc_vector, acc_step = recon(
            t0, t1, acc_vector, acc_step)

    def test_acc_step_for_impact_at_start(self):
        j = 0
        tj = self.times[j]
        tj_next = self.times[j+1]
        t = (tj + tj_next) / 2.
        self._recon_acc_for_impact_and_acc_for_intensity(t)

    def test_acc_step_for_impact(self):
        j = np.random.randint(len(self.times) - 1)
        tj = self.times[j]
        tj_next = self.times[j+1]
        t = (tj + tj_next) / 2.
        self._recon_acc_for_impact_and_acc_for_intensity(t)

    def _recon_acc_for_impact_and_acc_for_intensity(self, t):
        #         print('.....Recon accumulator for intensity and accumulator for impact....')
        #         print(f't: {t}')
        acc_full = np.zeros_like(self.alphas)
        for x1 in range(self.number_of_states):
            acc = np.zeros_like(self.alphas)
            intensities._acc_step_recon(
                0,
                x1,
                .0,
                t,
                self.betas,
                self.times,
                self.events,
                self.states,
                acc
            )
            acc_full[:, x1, :] = acc[:, x1, :]
        acc_imp = np.zeros_like(self.alphas)
        acc_imp = intensities._acc_step_for_impact_recon(
            .0,
            t,
            self.betas,
            self.times,
            self.events,
            self.states,
            acc_imp
        )
#         print(f'acc_full:\n{acc_full}')
#         print(f'acc_imp:\n{acc_imp}')
        rtol = 1e-7
        atol = 1e-12
        is_close = np.isclose(
            acc_full, acc_imp, rtol=rtol, atol=atol)
        error_msg = 'The accumulator for the impact intensity does not reconcile with the full accumulator\n'
        error_msg += f'acc_full:\n{acc_full}\n\n'
        error_msg += f'acc_imp:\n{acc_imp}\n'
        error_msg += '\nNon-reconciled entries:\n'
        error_msg += f'acc_full:\n{acc_full[~is_close]}\n\n'
        error_msg += f'acc_imp:\n{acc_imp[~is_close]}\n'

        self.assertTrue(
            np.allclose(
                acc_full,
                acc_imp,
                rtol=rtol,
                atol=atol,
            ),
            error_msg
        )

    def _direct_impact_sensitivity_test(
            self,
            direction,
    ):
        j = len(self.times) // 2
        t_j = self.times[j]
        t_j_next = self.times[j+1]
        tau = self.times[-1]
        s = (t_j + t_j_next) / 2.
        t = (s + t_j_next) / 2.
        self.assertEqual(
            intensities.direct_impact_spot_intensity(
                direction,
                t_j_next,
                t_j,
                self.phis,
                self.nus,
                self.alphas,
                self.betas,
                self.times,
                self.events,
                self.states
            ),
            0.,
            'Direct impact after execution is not zero'
        )
        impact_s = intensities.direct_impact_spot_intensity(
            direction,
            s,
            tau,
            self.phis,
            self.nus,
            self.alphas,
            self.betas,
            self.times,
            self.events,
            self.states
        )
        impact_t = intensities.direct_impact_spot_intensity(
            direction,
            t,
            tau,
            self.phis,
            self.nus,
            self.alphas,
            self.betas,
            self.times,
            self.events,
            self.states
        )
        # print(f'impact_s: {impact_s}')
        # print(f'impact_t: {impact_t}')
        self.assertLessEqual(
            impact_t,
            impact_s,
            'Direct impact is increasing between arrival times'
        )
        low_betas = np.array(self.betas, copy=True)
        low_betas[1:, :, 0] /= 2.
        impact_t_low_betas = intensities.direct_impact_spot_intensity(
            direction,
            t,
            tau,
            self.phis,
            self.nus,
            self.alphas,
            low_betas,
            self.times,
            self.events,
            self.states
        )
        # print(f'impact_t: {impact_t}')
        # print(f'impact_t_low_betas: {impact_t_low_betas}')
        error_msg = 'Decreasing the decay rate of the agent does not decrease the direct impact\n'
        error_msg += f'impact_t: {impact_t}\n'
        error_msg += f'impact_t_low_betas: {impact_t}\n'
        assertion = self.assertLess if impact_t > 0. else self.assertLessEqual
        assertion(
            impact_t,
            impact_t_low_betas,
            error_msg
        )
        phis_ = np.array(self.phis, copy=True)
        if direction < 0:
            phis_[:, 0, :self.number_of_states // 3] = 0.
        else:
            phis_[:, 0, -(self.number_of_states//3):] = 0.
        mass = np.sum(phis_[:, 0, :], axis=1, keepdims=True)
        if not np.all(mass > 0.):
            print(
                'Warning: some of the masses of transitions probabilities associated with event 0 are null')
            return
        phis_[:, 0, :] /= mass
        impact_t_zero_phis = intensities.direct_impact_spot_intensity(
            direction,
            t,
            tau,
            phis_,
            self.nus,
            self.alphas,
            self.betas,
            self.times,
            self.events,
            self.states
        )
        # print(f'impact_t_zero_phis: {impact_t_zero_phis}')
        self.assertLess(
            np.abs(impact_t_zero_phis),
            1e-9,
        )

    def test_seller_direct_impact_sensitivity(self):
        self._direct_impact_sensitivity_test(direction=-1)

    def test_buyer_direct_impact_sensitivity(self):
        self._direct_impact_sensitivity_test(direction=1)

    def _indirect_impact_sensitivity_test(
            self,
            direction,
    ):
        j = len(self.times) // 2
        t_j = self.times[j]
        t_j_next = self.times[j+1]
        tau = self.times[-1]
        s = (t_j + t_j_next) / 2.
        t = (s + t_j_next) / 2.
        u = (t_j + self.times[int(1.5 * j)]) / 2.
        impact_s = intensities.indirect_impact_spot_intensity(
            direction,
            s,
            self.phis,
            self.nus,
            self.alphas,
            self.betas,
            self.times,
            self.events,
            self.states
        )
        impact_t = intensities.indirect_impact_spot_intensity(
            direction,
            t,
            self.phis,
            self.nus,
            self.alphas,
            self.betas,
            self.times,
            self.events,
            self.states
        )
#         print(f'impact_s: {impact_s}')
#         print(f'impact_t: {impact_t}')
        self.assertLessEqual(
            abs(impact_t),
            abs(impact_s),
            'Absolute Indirect impact is increasing between arrival times'
        )
        eps = 1e-8
        phis = eps + np.array(self.phis, copy=True)
        symphi = np.array(phis, copy=True)
        dx3 = self.number_of_states // 3
        inflationary_mass = np.sum(
            phis[:, 1:, -dx3:],
            axis=2,
            keepdims=True)
        deflationary_mass = np.sum(
            phis[:, 1:, :dx3],
            axis=2,
            keepdims=True)
        dm = np.array(deflationary_mass, copy=True)
        im = np.array(inflationary_mass, copy=True)
        symphi[:, 1:, -dx3:] = (
            0.5 * (inflationary_mass + deflationary_mass) *
            symphi[:, 1:, -dx3:] / im
        )
        symphi[:, 1:, :dx3] = (
            0.5 * (inflationary_mass + deflationary_mass) *
            symphi[:, 1:, :dx3] / dm
        )

        sym_inflationary_mass = np.sum(
            symphi[:, 1:, -dx3:],
            axis=2,
            keepdims=True)
        sym_deflationary_mass = np.sum(
            symphi[:, 1:, :dx3],
            axis=2,
            keepdims=True)
#         print(f'Symmetric Deflationary masses: \n {sym_deflationary_mass}')
#         print(f'Symmetric Inflationary masses: \n {sym_inflationary_mass}')
        self.assertTrue(
            np.allclose(
                sym_inflationary_mass,
                sym_deflationary_mass,
                atol=1e-5,
                rtol=1e-3,
            )
        )
        sym_impact_t = intensities.indirect_impact_spot_intensity(
            direction,
            t,
            symphi,
            self.nus,
            self.alphas,
            self.betas,
            self.times,
            self.events,
            self.states
        )
        sym_impact_u = intensities.indirect_impact_spot_intensity(
            direction,
            u,
            symphi,
            self.nus,
            self.alphas,
            self.betas,
            self.times,
            self.events,
            self.states
        )
#         print(f'sym_impact_t: {sym_impact_t}')
#         print(f'sym_impact_u: {sym_impact_u}')
        self.assertTrue(
            np.allclose(
                sym_impact_u,
                .0,
                atol=1e-12,
                rtol=1e-6,
            )
        )
        self.assertTrue(
            np.allclose(
                sym_impact_t,
                .0,
                atol=1e-12,
                rtol=1e-6,
            )
        )

    def test_buyer_indirect_impact_sensitivity(
            self):
        self._indirect_impact_sensitivity_test(
            1
        )

    def test_seller_indirect_impact_sensitivity(
            self):
        self._indirect_impact_sensitivity_test(
            -1
        )

    def test_seller_impact_trajectory_and_spot_impact(
            self):
        direction = -1
        half = len(self.times) // 2
        j = half + np.random.choice(half)
        tj = self.times[j]
        t = (self.times[j-1] + tj) / 2.
        tau = self.times[-1] + 1e-4
        dt = self.times[0] / 4.
        self._recon_impact_trajectory_and_spot_impact(
            direction,
            t,
            tau,
            dt
        )

    def test_buyer_impact_trajectory_and_spot_impact(
            self):
        direction = 1
        half = len(self.times) // 2
        j = half + np.random.choice(half)
        tj = self.times[j]
        t = (self.times[j-1] + tj) / 2.
        tau = self.times[-1] + 1e-4
        dt = self.times[0] / 4.
        self._recon_impact_trajectory_and_spot_impact(
            direction,
            t,
            tau,
            dt
        )

    def test_fine_seller_impact_trajectory_and_spot_impact_at_start_with_no_agent_event(
            self):
        direction = -1
        j = 1
        tj = self.times[j]
        t = (self.times[j-1] + tj) / 2.
        tau = self.times[-1] + 1e-4
        dt = .85 * self.times[0] / 2.
        events = np.array(self.events, copy=True)
        if events[0] == 0:
            events[0] = np.random.choice(
                list(range(1, self.number_of_event_types)))
        self._recon_impact_trajectory_and_spot_impact(
            direction,
            t,
            tau,
            dt,
            events,
        )

    def test_fine_seller_impact_trajectory_and_spot_impact_at_start_with_agent_event(
            self):
        direction = -1
        j = 1
        tj = self.times[j]
        t = (self.times[j-1] + tj) / 2.
        tau = self.times[-1] + 1e-4
        dt = .85 * self.times[0] / 2.
        events = np.array(self.events, copy=True)
        events[0] = 0
        self._recon_impact_trajectory_and_spot_impact(
            direction,
            t,
            tau,
            dt,
            events,
        )

    def test_fine_buyer_impact_trajectory_and_spot_impact_at_start_with_agent_event(
            self):
        direction = 1
        j = 1
        tj = self.times[j]
        t = (self.times[j-1] + tj) / 2.
        tau = self.times[-1] + 1e-4
        dt = .85 * self.times[0] / 2.
        events = np.array(self.events, copy=True)
        events[0] = 0
        self._recon_impact_trajectory_and_spot_impact(
            direction,
            t,
            tau,
            dt,
            events,
        )

    def test_impact_trajectory_and_spot_impact_at_first_10(
            self):
        direction = int(np.random.choice([-1, 1]))
        j = max(1, -1 + min(11, len(self.times)))
        tj = self.times[j]
        t = (self.times[j-1] + tj) / 2.
        tau = self.times[-1] + 1e-4
        dt = self.times[j + 1]
        events = self.events
        self._recon_impact_trajectory_and_spot_impact(
            direction,
            t,
            tau,
            dt,
            events,
        )

    def test_fine_impact_trajectory_and_spot_impact_at_first_10(
            self):
        direction = int(np.random.choice([-1, 1]))
        j = max(1, -1 + min(11, len(self.times)))
        tj = self.times[j]
        t = (self.times[j-1] + tj) / 2.
        tau = self.times[-1] + 1e-4
        dt = float(
            np.random.uniform(low=.5, high=.95) * self.times[0] / 2.
        )
        events = self.events
        self._recon_impact_trajectory_and_spot_impact(
            direction,
            t,
            tau,
            dt,
            events,
        )

    def test_seller_impact_trajectory_and_spot_impact_at_start_with_no_agent_event(
            self):
        direction = -1
        j = 1
        tj = self.times[j]
        t = (self.times[j-1] + tj) / 2.
        tau = self.times[-1] + 1e-4
        dt = self.times[0] + 1.
        events = np.array(self.events, copy=True)
        if events[0] == 0:
            events[0] = np.random.choice(
                list(range(1, self.number_of_event_types)))
        self._recon_impact_trajectory_and_spot_impact(
            direction,
            t,
            tau,
            dt,
            events,
        )

    def test_seller_impact_trajectory_and_spot_impact_at_start_with_agent_event(
            self):
        direction = -1
        j = 1
        tj = self.times[j]
        t = (self.times[j-1] + tj) / 2.
        tau = self.times[-1] + 1e-4
        dt = self.times[0] + 1.
        events = np.array(self.events, copy=True)
        events[0] = 0
        self._recon_impact_trajectory_and_spot_impact(
            direction,
            t,
            tau,
            dt,
            events,
        )

    def test_seller_impact_trajectory_and_spot_impact_at_second_start_with_agent_event(
            self):
        direction = -1
        j = 2
        tj = self.times[j]
        t = (self.times[j-1] + tj) / 2.
        tau = self.times[-1] + 1e-4
        dt = self.times[j] + 1.
        events = np.array(self.events, copy=True)
        events[0] = 0
        self._recon_impact_trajectory_and_spot_impact(
            direction,
            t,
            tau,
            dt,
            events,
        )

    def test_seller_impact_trajectory_and_spot_impact_at_second_start(
            self):
        direction = -1
        j = 2
        tj = self.times[j]
        t = (self.times[j-1] + tj) / 2.
        tau = self.times[-1] + 1e-4
        dt = self.times[j] + 1.
        events = self.events
        self._recon_impact_trajectory_and_spot_impact(
            direction,
            t,
            tau,
            dt,
            events,
        )

    def _recon_impact_trajectory_and_spot_impact(
            self,
            direction,
            t,
            tau,
            dt,
            events=None
    ):
        if events is None:
            events = self.events
        di = intensities.direct_impact_spot_intensity(
            direction,
            t,
            tau,
            self.phis,
            self.nus,
            self.alphas,
            self.betas,
            self.times,
            events,
            self.states,
        )
        ii = intensities.indirect_impact_spot_intensity(
            direction,
            t,
            self.phis,
            self.nus,
            self.alphas,
            self.betas,
            self.times,
            events,
            self.states,
        )
        impact = intensities.impact(
            direction,
            t,
            tau,
            dt,
            self.phis,
            self.nus,
            self.alphas,
            self.betas,
            self.times,
            events,
            self.states,
        )
        tell_time = f't: {t}\n'
        tell_direct_impact = f'Direct impact: {di}\n'
        tell_indirect_impact = f'Indirect impact: {ii}\n'
        tell_last_point = f'len(impact): {len(impact)}; impact[-5:, :]:\n {impact[-5:, :]}\n'
#         print(tell_time)
#         print(tell_direct_impact)
#         print(tell_indirect_impact)
#         print(tell_last_point)
        error_msg = '\nImpact from trajectory does not reconcile with spot impact\n'
        error_msg += tell_time
        error_msg += tell_direct_impact + tell_indirect_impact
        error_msg += tell_last_point
        self.assertTrue(
            np.allclose(
                np.array(
                    [t, di, ii]
                ),
                impact[-1, :],
                rtol=1e-4,
                atol=1e-8,
            ),
            error_msg
        )

    def test_after_sell_execution_decrease(self):
        direction = -1
        t = self.times[-1]
        tau = self.times[len(self.times) // 2]
        self._test_after_execution_decrease(
            direction,
            t,
            tau
        )

    def test_after_buy_execution_decrease(self):
        direction = 1
        t = self.times[-1]
        tau = self.times[len(self.times) // 2]
        self._test_after_execution_decrease(
            direction,
            t,
            tau
        )

    def _test_after_execution_decrease(
            self,
            direction,
            t,
            tau
    ):
        dt = .25 * self.times[0]
        impact = intensities.impact(
            direction,
            t,
            tau,
            dt,
            self.phis,
            self.nus,
            self.alphas,
            self.betas,
            self.times,
            self.events,
            self.states,
        )
        t1 = bisect.bisect_right(impact[:, 0], tau)
        self.assertTrue(
            np.allclose(
                np.diff(impact[t1:, 1]),
                0.
            ),
            'Intensity of direct impact is not null after execution',
        )


if __name__ == '__main__':
    unittest.main()



from typing import Optional, Sequence
import numpy as np
import math
import copy
import bisect
import scipy.optimize as opt
from . import hybrid_hawkes_exp_cython as cy


class HybridHawkesExp:
    """
    This class implements state-dependent Hawkes processes with exponential kernels, a subclass of hybrid marked point
    processes.
    The main features it provides include simulation and statistical inference (estimation).

    :type number_of_event_types: int
    :param number_of_event_types: number of different event types.
    :type number_of_states: int
    :param number_of_states: number of possible states.
    :type events_labels: list of strings
    :param events_labels: names of the different event types.
    :type states_labels: list of strings
    :param states_labels: names of the possible states.
    """

    def __init__(self, number_of_event_types, number_of_states, events_labels, states_labels):
        """
        Initialises an instance.

        :type number_of_event_types: int
        :param number_of_event_types: number of different event types.
        :type number_of_states: int
        :param number_of_states: number of possible states.
        :type events_labels: list of strings
        :param events_labels: names of the different event types.
        :type states_labels: list of strings
        :param states_labels: names of the possible states.
        """
        self.number_of_event_types = number_of_event_types
        self.number_of_states = number_of_states
        self.events_labels = events_labels
        self.states_labels = states_labels
        self.transition_probabilities = np.zeros(
            (number_of_states, number_of_event_types, number_of_states))
        self.base_rates = np.zeros(number_of_event_types)
        self.impact_coefficients = np.zeros(
            (number_of_event_types, number_of_states, number_of_event_types))
        self.decay_coefficients = np.zeros(
            (number_of_event_types, number_of_states, number_of_event_types))
        self.impact_decay_ratios = np.zeros(
            (number_of_event_types, number_of_states, number_of_event_types))

    def set_transition_probabilities(self, transition_probabilities):
        r"""
        Fixes the transition probabilities :math:`\phi` of the state-dependent Hawkes process.
        The are used to :py:meth:`~mpoints.hybrid_hawkes_exp.HybridHawkesExp.simulate` and
        :py:meth:`~mpoints.hybrid_hawkes_exp.HybridHawkesExp.compute_total_residuals`.

        :type transition_probabilities: 3D numpy array
        :param transition_probabilities: shape should be :math:`(d_x, d_e,d_x)` where :math:`d_e` and :math:`d_x`
                                         are the number of event types and states, respectively.
                                         The entry :math:`i, j, k` is the probability of going from state :math:`i`
                                         to state :math:`k` when an event of type :math:`j` occurs.
        :return:
        """
        'Raise ValueError if the given parameters do not have the right shape'
        if np.shape(transition_probabilities) != (self.number_of_states, self.number_of_event_types,
                                                  self.number_of_states):
            raise ValueError(
                'given transition probabilities have incorrect shape')
        self.transition_probabilities = copy.copy(transition_probabilities)

    def set_hawkes_parameters(self, base_rates, impact_coefficients, decay_coefficients):
        r"""
        Fixes the parameters :math:`(\nu, \alpha, \beta)` that define the intensities (arrival rates) of events.
        The are used in
        :py:meth:`~mpoints.hybrid_hawkes_exp.HybridHawkesExp.simulate`,
        :py:meth:`~mpoints.hybrid_hawkes_exp.HybridHawkesExp.compute_events_residuals`
        and :py:meth:`~mpoints.hybrid_hawkes_exp.HybridHawkesExp.compute_total_residuals`.

        :type base_rates: 1D numpy array
        :param base_rates: one base rate :math:`\nu_e` per event type :math:`e`.
        :type impact_coefficients: 3D numpy array
        :param impact_coefficients: the alphas :math:`\alpha_{e'xe}`.
        :type decay_coefficients: 3D numpy array
        :param decay_coefficients: the betas :math:`\beta_{e'xe}`.
        :return:
        """
        'Raise ValueError if the given parameters do not have the right shape'
        if np.shape(base_rates) != (self.number_of_event_types,):
            raise ValueError('given base rates have incorrect shape')
        if np.shape(impact_coefficients) != (self.number_of_event_types, self.number_of_states,
                                             self.number_of_event_types):
            raise ValueError('given impact coefficients have incorrect shape')
        if np.shape(decay_coefficients) != (self.number_of_event_types, self.number_of_states,
                                            self.number_of_event_types):
            raise ValueError('given decay coefficients have incorrect shape')
        self.base_rates = copy.copy(base_rates)
        self.impact_coefficients = copy.copy(impact_coefficients)
        self.decay_coefficients = copy.copy(decay_coefficients)
        self.impact_decay_ratios = np.divide(
            impact_coefficients, decay_coefficients)

    @staticmethod
    def kernel_at_time(time, alpha, beta):
        r"""
        Evaluates the kernel of the model at the given time with the given parameters.

        :type time: float
        :param time: the positive time :math:`t`.
        :type alpha: float
        :param alpha: a non-negative :math:`\alpha`.
        :type beta: float
        :param beta: a positive :math:`\beta`.
        :rtype: float
        :return: :math:`\alpha\exp(-\beta t)`.
        """
        return alpha * np.exp(- np.multiply(time, beta))

    'Functions that estimate the model parameters'

    def estimate_transition_probabilities(self, events, states):
        r"""
        Estimates the transition probabilities :math:`\phi` of the state process from the data.
        This method returns the maximum likelihood estimate.
        One can prove that it coincides with the empirical transition probabilities.

        :type events: 1D array of int
        :param events: the sequence of event types, `events[n]` is the event type of the `n` th event.
        :type states: 1D array of int
        :param states: the sequence of states, `states[n]` is the new state of the system following the `n` th event.
        :rtype: 3D array
        :return: the estimated transition probabilities :math:`\phi`.
        """
        result = np.zeros(
            (self.number_of_states, self.number_of_event_types, self.number_of_states))
        count_of_states_events = np.zeros(
            (self.number_of_states, self.number_of_event_types))
        for n in range(1, len(events)):
            event = events[n]
            state_before = states[n - 1]
            state_after = states[n]
            count_of_states_events[state_before, event] += 1
            result[state_before, event, state_after] += 1
        for x1 in range(self.number_of_states):
            for e in range(self.number_of_event_types):
                size = count_of_states_events[x1, e]
                if size > 0:
                    for x2 in range(self.number_of_states):
                        result[x1, e, x2] /= size
                else:
                    message = 'Warning: Transition probabilities from state ' + \
                        str(x1)
                    message += ' when events of type ' + \
                        str(e) + ' occur cannot be estimated because'
                    message += ' events of this type never occur this state'
                    print(message)
        return result

    def estimate_hawkes_parameters(self, times, events, states, time_start, time_end, maximum_number_of_iterations=2000,
                                   method='TNC', parameters_lower_bound=10**(-6), parameters_upper_bound=None,
                                   given_guesses=[], number_of_random_guesses=1,
                                   min_decay_coefficient=0.5, max_decay_coefficient=100, parallel_estimation=True):
        r"""
        Estimates the parameters of the intensities (arrival rates) of events, i.e., :math:`(\nu, \alpha, \beta)`.
        Estimation if performed via maximum likelihood. This method uses the `scipy.minimize` library.

        :type times: 1D numpy array of float
        :param times: the times at which events occur.
        :type events: 1D numpy array of int
        :param events: the sequence of event types, `events[n]` is the event type of the `n` th event.
        :type states: 1D numpy array of int
        :param states: the sequence of states, `states[n]` is the new state of the system following the `n` th event.
        :type time_start: float
        :param time_start: the time at which we consider that the process started, prior times are treated as an
                           initial condition.
        :type time_end: float
        :param time_end: the time at which we stopped to record the process.
        :type maximum_number_of_iterations: int
        :param maximum_number_of_iterations:  will be passed to the `maxiter` argument in `scipy.minimize`.
                                              Depending on `method`, it is the maximum number of iterations or
                                              function evaluations.
        :type method: string
        :param method: the optimisation method used in `scipy.minimize`.
        :type parameters_lower_bound: float
        :param parameters_lower_bound: lower bound on all the parameters.
        :type parameters_upper_bound: float
        :param parameters_upper_bound: upper bound on all the parameters.
        :type given_guesses: list of 1D numpy array
        :param given_guesses: every member `x` is an initial guess on the parameters.
                              For every `x`, we attempt to maximise the likelihood starting from `x`.
                              One can go from `x` to :math:`(\nu, \alpha, \beta)` and vice versa using
                              :py:meth:`~mpoints.hybrid_hawkes_exp.HybridHawkesExp.array_to_parameters`
                              and :py:meth:`~mpoints.hybrid_hawkes_exp.HybridHawkesExp.parameters_to_array`.
                              We retain the solution that gives the highest likelihood.
        :type number_of_random_guesses: int
        :param number_of_random_guesses: the method can also generate random initial guesses.
        :type min_decay_coefficient: numpy array or float
        :param min_decay_coefficient: defines how a random guess is generated.
        :type max_decay_coefficient: numpy array of float
        :param max_decay_coefficient: a random guess on :math:`\beta_{e'xe}` is generated uniformly in the interval
                                      [`min_decay_coefficient[e',x,e]`, `max_decay_coefficient[e',x.e]`] but on a
                                      logarithmic scale.
        :type parallel_estimation: boolean
        :param parallel_estimation: the MLE problem can be decomposed into :math:`d_e` independent optimisation
                                    problems, where :math:`d_e` is the number of event types. When True, each problem
                                    is solved independently. In this case, the limit on the number of iterations
                                    or function evaluations is applied independently to each sub-problem.
        :rtype: scipy.optimize.OptimizerResult, 1D numpy array, string
        :return: The first object is the optimisation result and contains the maximum likelihood estimate along with
                 additional information on the optimisation routine. The second object contains the initial guess
                 that resulted in the highest likelihood after running the optimisation procedure.
                 The third object indicates the nature of this initial guess ('random' or 'given').
        """

        'Generate additional random guesses of the parameters'
        guesses = copy.copy(given_guesses)
        # if a scalar was given instead of a matrix
        if np.shape(min_decay_coefficient) == ():
            min_decay_coefficients = min_decay_coefficient * np.ones((self.number_of_event_types, self.number_of_states,
                                                                      self.number_of_event_types))
        # if a scalar was given instead of a matrix
        if np.shape(max_decay_coefficient) == ():
            max_decay_coefficients = max_decay_coefficient * np.ones((self.number_of_event_types, self.number_of_states,
                                                                      self.number_of_event_types))
        if number_of_random_guesses > 0:
            'Compute the average intensities'
            average_intensities = np.zeros(self.number_of_event_types)
            for n in range(len(times)):
                e = events[n]
                average_intensities[e] += 1
            average_intensities = np.divide(
                average_intensities, time_end - time_start)
            for n in range(number_of_random_guesses):
                'Base rates'
                guess_base_rates = np.zeros(self.number_of_event_types)
                for e in range(self.number_of_event_types):
                    guess_base_rates[e] = average_intensities[e] / 2
                'Decay coefficients'
                guess_decay_coefficients = np.zeros((self.number_of_event_types, self.number_of_states,
                                                     self.number_of_event_types))
                for e1 in range(self.number_of_event_types):
                    for x in range(self.number_of_states):
                        for e2 in range(self.number_of_event_types):
                            u_min = math.log10(
                                min_decay_coefficients[e1, x, e2])
                            u_max = math.log10(
                                max_decay_coefficients[e1, x, e2])
                            u = np.random.uniform(u_min, u_max)
                            beta = 10 ** u
                            guess_decay_coefficients[e1, x, e2] = beta
                'Impact coefficients'
                guess_impact_coefficients = np.zeros((self.number_of_event_types, self.number_of_states,
                                                      self.number_of_event_types))
                for e1 in range(self.number_of_event_types):
                    for x in range(self.number_of_states):
                        for e2 in range(self.number_of_event_types):
                            u = np.random.uniform(0, 1)
                            alpha = u * guess_decay_coefficients[e1, x, e2]
                            guess_impact_coefficients[e1, x, e2] = alpha
                'Save the random guess to the list of guesses'
                g = HybridHawkesExp.parameters_to_array(guess_base_rates, guess_impact_coefficients,
                                                        guess_decay_coefficients)
                guesses.append(g)

        'For each initial guess, apply the optimizer'
        if not parallel_estimation:
            optimal_results = []
            for g in guesses:
                dimension = self.number_of_event_types + 2 * \
                    self.number_of_states * self.number_of_event_types ** 2
                bounds = [(parameters_lower_bound,
                           parameters_upper_bound)] * dimension
                'Define the minus likelihood and gradient functions'
                def likelihood_minus(parameters):
                    result = - self.log_likelihood_of_events(parameters, times, events, states,
                                                             time_start, time_end)
                    return result

                def gradient_of_likelihood_minus(parameters):
                    result = - \
                        self.gradient(parameters, times, events,
                                      states, time_start, time_end)
                    return result
                o = opt.minimize(likelihood_minus, g, method=method,
                                 bounds=bounds, jac=gradient_of_likelihood_minus,
                                 options={'maxiter': maximum_number_of_iterations})
                optimal_results.append(o)
            'Look for the solution that gives the highest log-likelihood'
            index_of_best_result = 0
            log_likelihood_minus = optimal_results[0].fun
            for i in range(1, len(optimal_results)):
                current_log_likelihood_minus = optimal_results[i].fun
                if current_log_likelihood_minus < log_likelihood_minus:
                    index_of_best_result = i
                    log_likelihood_minus = current_log_likelihood_minus
            best_initial_guess = guesses[index_of_best_result]
            kind_of_best_initial_guess = ''
            if index_of_best_result < len(given_guesses):
                kind_of_best_initial_guess += 'given'
            elif index_of_best_result - len(given_guesses) < number_of_random_guesses:
                kind_of_best_initial_guess += 'random'
            'Return the OptimizeResult instance that gives the biggest likelihood'
            return optimal_results[index_of_best_result], best_initial_guess, kind_of_best_initial_guess
        else:
            dimension = 1 + 2 * self.number_of_states * self.number_of_event_types
            bounds = [(parameters_lower_bound,
                       parameters_upper_bound)] * dimension
            opt_nus = np.zeros(self.number_of_event_types)
            opt_alphas = np.zeros(
                (self.number_of_event_types, self.number_of_states, self.number_of_event_types))
            opt_betas = np.zeros(
                (self.number_of_event_types, self.number_of_states, self.number_of_event_types))
            best_guess_nu = np.zeros(self.number_of_event_types)
            best_guess_alphas =\
                np.zeros((self.number_of_event_types,
                         self.number_of_states, self.number_of_event_types))
            best_guess_betas = np.zeros(
                (self.number_of_event_types, self.number_of_states, self.number_of_event_types))
            success = True
            successes = []
            status = -999
            statuses = []
            message = 'Multiple messages because parallel estimation'
            messages = []
            fun = 0
            jacs = []
            hesss = []
            nfev = 0
            nit = 0
            kinds_of_best_initial_guesses = ''
            for e in range(self.number_of_event_types):
                'Define the minus likelihood and gradient functions'
                def likelihood_minus(parameters):
                    result = - self.log_likelihood_of_events_partial(e, parameters, times, events, states,
                                                                     time_start, time_end)
                    return result

                def gradient_of_likelihood_minus(parameters):
                    result = - \
                        self.gradient_partial(
                            e, parameters, times, events, states, time_start, time_end)
                    return result
                'For each initial guess, optimise likelihood'
                optimal_results = []
                for g in guesses:
                    guess_nus, guess_alphas, guess_betas = self.array_to_parameters(g, self.number_of_event_types,
                                                                                    self.number_of_states,
                                                                                    self.number_of_event_types)
                    g_partial = self.parameters_to_array(guess_nus[e:e+1],
                                                         guess_alphas[:,
                                                                      :, e:e+1],
                                                         guess_betas[:, :, e:e+1])
                    o = opt.minimize(likelihood_minus, g_partial, method=method,
                                     bounds=bounds, jac=gradient_of_likelihood_minus,
                                     options={'maxiter': maximum_number_of_iterations})
                    optimal_results.append(o)
                'Look for the solution that gives the highest log-likelihood'
                index_of_best_result = 0
                log_likelihood_minus = optimal_results[0].fun
                for i in range(1, len(optimal_results)):
                    current_log_likelihood_minus = optimal_results[i].fun
                    if current_log_likelihood_minus < log_likelihood_minus:
                        index_of_best_result = i
                        log_likelihood_minus = current_log_likelihood_minus
                best_initial_guess = guesses[index_of_best_result]
                kind_of_best_initial_guess = ''
                if index_of_best_result < len(given_guesses):
                    kind_of_best_initial_guess += 'given'
                elif index_of_best_result - len(given_guesses) < number_of_random_guesses:
                    kind_of_best_initial_guess += 'random'
                o = optimal_results[index_of_best_result]
                # Save the kind of best initial guess for this event type
                kinds_of_best_initial_guesses += kind_of_best_initial_guess + ' '
                # Save optimal parameters
                v, a, b = self.array_to_parameters(
                    o.x, self.number_of_event_types, self.number_of_states, 1)
                opt_nus[e:e+1] = v
                opt_alphas[:, :, e:e+1] = a
                opt_betas[:, :, e:e+1] = b
                # Save best initial guess
                v, a, b = self.array_to_parameters(best_initial_guess, self.number_of_event_types,
                                                   self.number_of_states, self.number_of_event_types)
                best_guess_nu[e] = v[e]
                best_guess_alphas[:, :, e] = a[:, :, e]
                best_guess_betas[:, :, e] = b[:, :, e]
                # Save optimiser information
                successes.append(o.success)
                statuses.append(o.status)
                messages.append(o.message)
                if success and not o.success:
                    success = False
                    status = o.status
                fun += o.fun
                jacs.append(o.jac)
                nfev += o.nfev
                nit += o.nit
            'Aggregate the Optimize Results into a single one'
            best_initial_guess = self.parameters_to_array(
                best_guess_nu, best_guess_alphas, best_guess_betas)
            o = opt.OptimizeResult()
            x = self.parameters_to_array(opt_nus, opt_alphas, opt_betas)
            o['x'] = x
            o['success'] = success
            o['successes'] = successes
            o['status'] = status
            o['statuses'] = statuses
            o['message'] = message
            o['messages'] = messages
            o['fun'] = fun
            o['jacs'] = jacs
            o['hesss'] = hesss
            o['nfev'] = nfev
            o['nit'] = nit
            return o, best_initial_guess, kinds_of_best_initial_guesses

    'Specification testing and simulation'

    def compute_events_residuals(self, times, events, states, time_start, initial_partial_sums=0):
        r"""
        Computes the events residuals :math:`r^e_n` defined by

        .. math::

            r^e_n := \int_{t^e_{n-1}}^{t^e_n} \lambda_e (t)dt,

        where :math:`t^e_n` is the time when the `n` th event of type `e` occurred.
        The methods wraps a C implementation that was obtained via Cython.

        :type times: 1D numpy array of float
        :param times: the times at which events occur.
        :type events: 1D numpy array of int
        :param events: the sequence of event types, `events[n]` is the event type of the `n` th event.
        :type states: 1D numpy array of int
        :param states: the sequence of states, `states[n]` is the new state of the system following the `n` th event.
        :type time_start: float
        :param time_start: the time at which we consider that the process started, prior times are treated as an
                           initial condition.
        :type initial_partial_sums: 3D numpy array
        :param initial_partial_sums: the initial condition can also be given implicitly via the partial sums
                                     :math:`S_{e',x,e}(-\infty, \mbox{time_start}]`.
        :rtype: list of 1D numpy arrays
        :return: the `e` th element of the list is the sequence :math:`(r^e_n)` corresponding to the event type `e`.
        """
        # Check if no initial partial sums if given
        s = np.zeros((self.number_of_event_types,
                     self.number_of_states, self.number_of_event_types))
        if len(np.shape(initial_partial_sums)) != 0:
            s = initial_partial_sums
            s = np.divide(s, self.decay_coefficients)
        return cy.compute_events_residuals(self.base_rates,
                                           self.impact_coefficients,
                                           self.decay_coefficients,
                                           self.number_of_event_types,
                                           self.number_of_states,
                                           times,
                                           events,
                                           states,
                                           time_start,
                                           s)

    def compute_total_residuals(self, times, events, states, time_start, initial_partial_sums=0,
                                initial_state=0):
        r"""

        Computes the total residuals :math:`r^{ex}_n` defined by

        .. math::

            r^{ex}_n := \int_{t^{ex}_{n-1}}^{t^{ex}_n}\phi_{e}(X(t),x) \lambda_e (t)dt,

        where :math:`t^{ex}_n` is the time when the `n` th event of type `e` after which the state is `x` occurred.
        The methods wraps a C implementation that was obtained via Cython.

        :type times: 1D numpy array of float
        :param times: the times at which events occur.
        :type events: 1D numpy array of int
        :param events: the sequence of event types, `events[n]` is the event type of the `n` th event.
        :type states: 1D numpy array of int
        :param states: the sequence of states, `states[n]` is the new state of the system following the `n` th event.
        :type time_start: float
        :param time_start: the time at which we consider that the process started, prior times are treated as an
                           initial condition.
        :type initial_partial_sums: 3D numpy array
        :param initial_partial_sums: the initial condition can also be given implicitly via the partial sums
                                     :math:`S_{e',x,e}(-\infty, \mbox{time_start}]`.
        :type initial_state: int
        :param initial_state: if there are no event times before `time_start`, this is used as the initial state.
        :rtype: list of 1D numpy arrays
        :return: the sequence :math:`(r^{ex}_n)` is the `x` + `e` * `number_of_states` th element in the list.
        """
        # Check if no initial partial sums if given
        s = np.zeros((self.number_of_event_types,
                     self.number_of_states, self.number_of_event_types))
        if len(np.shape(initial_partial_sums)) != 0:
            s = initial_partial_sums
            s = np.divide(s, self.decay_coefficients)
        return cy.compute_total_residuals(self.transition_probabilities,
                                          self.base_rates,
                                          self.impact_coefficients,
                                          self.decay_coefficients,
                                          self.number_of_event_types,
                                          self.number_of_states,
                                          times,
                                          events,
                                          states,
                                          time_start,
                                          s,
                                          initial_state)

    def simulate(self,
                 time_start: float,
                 time_end: float,
                 initial_condition_times: Optional[Sequence] = None,
                 initial_condition_events: Optional[Sequence] = None,
                 initial_condition_states: Optional[Sequence] = None,
                 initial_partial_sums: Optional[np.array] = None,
                 initial_state: int = 0,
                 max_number_of_events: int = 10**6):
        """
        Simulates a sample path of the state-dependent Hawkes process.
        The methods wraps a C implementation that was obtained via Cython.

        :type time_start: float
        :param time_start: time at which the simulation starts.
        :type time_end: float
        :param time_end: time at which the simulation ends.
        :type initial_condition_times: array
        :param initial_condition_times: times of events before and including `time_start`.
        :type initial_condition_events: array of int
        :param initial_condition_events: types of the events that occurred at `initial_condition_times`.
        :type initial_condition_states: array of int
        :param initial_condition_states: values of the state process just after the `initial_condition_times`.
        :type initial_partial_sums: 3D numpy array
        :param initial_partial_sums: the initial condition can also be given implicitly via the partial sums
                                     :math:`S_{e',x,e}(-\infty, \mbox{time_start}]`.
        :type initial_state: int
        :param initial_state: if there are no event times before `time_start`, this is used as the initial state.
        :type max_number_of_events: int
        :param max_number_of_events: the simulation stops when this number of events is reached
                                     (including the initial condition).
        :rtype: array, array of int, array of int
        :return: the times at which the events occur, their types and the values of the state process right after
                 each event. Note that these include the initial condition as well.
        """
        # Check if no initial partial sums if given
        if initial_partial_sums is None:
            s = np.zeros((self.number_of_event_types,
                         self.number_of_states, self.number_of_event_types))
        elif isinstance(initial_partial_sums, (int, float)):
            s = initial_partial_sums * np.ones((self.number_of_event_types,
                                                self.number_of_states,
                                                self.number_of_event_types), dtype=float)
        else:
            assert isinstance(initial_partial_sums, np.ndarray)
            assert initial_partial_sums.shape == (
                self.number_of_event_types, self.number_of_states, self.number_of_event_types)
            s = np.array(initial_partial_sums, dtype=float)
        # Convert the initial condition to np.arrays if required
        if initial_condition_times is None:
            initial_condition_times = np.array([], dtype=float)
        elif not isinstance(initial_condition_times, np.ndarray):
            initial_condition_times = np.asarray(
                initial_condition_times, dtype=float)
        if initial_condition_events is None:
            initial_condition_events = np.array([], dtype=int)
        elif not isinstance(initial_condition_events, np.ndarray):
            initial_condition_events = np.asarray(
                initial_condition_events, dtype=int)
        if initial_condition_states is None:
            initial_condition_states = np.array([], dtype=int)
        elif not isinstance(initial_condition_states, np.ndarray):
            initial_condition_states = np.asarray(
                initial_condition_states, dtype=int)
        return cy.simulate(
            self.number_of_event_types,
            self.number_of_states,
            self.base_rates,
            self.impact_coefficients,
            self.decay_coefficients,
            self.transition_probabilities,
            initial_condition_times,
            initial_condition_events,
            initial_condition_states,
            s,
            initial_state,
            time_start,
            time_end,
            max_number_of_events
        )

    'Likelihood and gradient'

    def log_likelihood_of_events(self, parameters, times, events, states, time_start, time_end):
        r"""
        Computes the log-likelihood of the observed times and event types under the assumption that they
        are the realisation of a state-dependent Hawkes process with the given parameters.
        The log-likelihood is given by

        .. math::

            l := \sum_{n : t_0 < t_n \leq T} \lambda_{e_n}(t_n) - \sum_e \int_{t_0}^T \lambda_e(t)dt,

        where :math:`(t_n)` and :math:`(e_n)` are the sequences of event times and event types, respectively.
        The method wraps a C implementation that was obtained via Cython.

        :type parameters: 1D numpy array
        :param parameters: the parameters :math:`(\nu, \alpha, \beta)` put into a single array.
                           Go from `parameters` to :math:`(\nu, \alpha, \beta)` and vice versa using
                           :py:meth:`~mpoints.hybrid_hawkes_exp.HybridHawkesExp.array_to_parameters`
                           and :py:meth:`~mpoints.hybrid_hawkes_exp.HybridHawkesExp.parameters_to_array`.
        :type times: 1D numpy array of float
        :param times: the times at which events occur.
        :type events: 1D numpy array of int
        :param events: the sequence of event types, `events[n]` is the event type of the `n` th event.
        :type states: 1D numpy array of int
        :param states: the sequence of states, `states[n]` is the new state of the system following the `n` th event.
        :type time_start: float
        :param time_start: :math:`t_0`, the time at which we consider that the process started, prior times are treated as an
                           initial condition.
        :type time_end: float
        :param time_end: :math:`T`, the time at which we stopped to record the process.
        :rtype: float
        :return: the log-likelihood :math:`l`.
        """
        number_of_event_types = self.number_of_event_types
        number_of_states = self.number_of_states
        base_rates, impact_coefficients, decay_coefficients = \
            HybridHawkesExp.array_to_parameters(
                parameters, self.number_of_event_types, self.number_of_states)
        return cy.log_likelihood_of_events(base_rates, impact_coefficients, decay_coefficients, number_of_event_types,
                                           number_of_states, times, events, states, float(time_start), float(time_end))

    def gradient(self, parameters, times, events, states, time_start, time_end):
        r"""
        Computes the gradient of the log-likelihood :math:`l` with respect to the
        parameters :math:`(\nu, \alpha, \beta)`.
        The method wraps a C implementation that was obtained via Cython.

        :type parameters: 1D numpy array
        :param parameters: the parameters :math:`(\nu, \alpha, \beta)` put into a single array.
                           Go from `parameters` to :math:`(\nu, \alpha, \beta)` and vice versa using
                           :py:meth:`~mpoints.hybrid_hawkes_exp.HybridHawkesExp.array_to_parameters`
                           and :py:meth:`~mpoints.hybrid_hawkes_exp.HybridHawkesExp.parameters_to_array`.
        :type times: 1D numpy array of float
        :param times: the times at which events occur.
        :type events: 1D numpy array of int
        :param events: the sequence of event types, `events[n]` is the event type of the `n` th event.
        :type states: 1D numpy array of int
        :param states: the sequence of states, `states[n]` is the new state of the system following the `n` th event.
        :type time_start: float
        :param time_start: :math:`t_0`, the time at which we consider that the process started, prior times are treated as an
                           initial condition.
        :type time_end: float
        :param time_end: :math:`T`, the time at which we stopped to record the process.
        :rtype: float
        :return: the gradient of the log-likelihood :math:`l`.
        """
        number_of_event_types = self.number_of_event_types
        number_of_states = self.number_of_states
        base_rates, impact_coefficients, decay_coefficients = \
            HybridHawkesExp.array_to_parameters(
                parameters, self.number_of_event_types, self.number_of_states)
        g_base_rates, g_impact_coefficients, g_decay_coefficients =\
            cy.gradient(base_rates, impact_coefficients, decay_coefficients, number_of_event_types,
                        number_of_states, times, events, states, float(time_start), float(time_end))
        return self.parameters_to_array(g_base_rates, g_impact_coefficients, g_decay_coefficients)

    def log_likelihood_of_events_partial(self, event_type, parameters, times, events, states, time_start, time_end):
        r"""
        Computes the log-likelihood of the arrival times of events of the given type under the assumption that they
        are the realisation of a state-dependent Hawkes process with the given parameters.
        These parameters include only those that govern :math:`\lambda_e`, the intensity of events of type `e`.
        For example, among the base rates :math:`\nu`, only :math:`\nu_e` should be contained in `parameters`.
        This partial log-likelihood is given by

        .. math::

            l_e := \sum_{n : t_0 < t^e_n \leq T} \lambda_{e}(t^e_n) - \int_{t_0}^T \lambda_e(t)dt,

        where :math:`(t^e_n)` are the arrival times of events of the given type `e`.
        The method wraps a C implementation that was obtained via Cython.

        :type event_type: int
        :param event_type: `e`, the event type for which we want to compute the partial log-likelihood :math:`l_e`.
        :type parameters: 1D numpy array
        :param parameters: only the parameters :math:`(\nu, \alpha, \beta)` that govern :math:`\lambda_e`  put into a single
                           array.
                           Go from `parameters` to :math:`(\nu, \alpha, \beta)` and vice versa using
                           :py:meth:`~mpoints.hybrid_hawkes_exp.HybridHawkesExp.array_to_parameters`
                           and :py:meth:`~mpoints.hybrid_hawkes_exp.HybridHawkesExp.parameters_to_array`.
        :type times: 1D numpy array of float
        :param times: the times at which events occur.
        :type events: 1D numpy array of int
        :param events: the sequence of event types, `events[n]` is the event type of the `n` th event.
        :type states: 1D numpy array of int
        :param states: the sequence of states, `states[n]` is the new state of the system following the `n` th event.
        :type time_start: float
        :param time_start: :math:`t_0`, the time at which we consider that the process started, prior times are treated as an
                           initial condition.
        :type time_end: float
        :param time_end: :math:`T`, the time at which we stopped to record the process.
        :rtype: float
        :return: the partial log-likelihood :math:`l_e`.
        """
        number_of_event_types = self.number_of_event_types
        number_of_states = self.number_of_states
        base_rate, impact_coefficients, decay_coefficients = \
            HybridHawkesExp.array_to_parameters(
                parameters, number_of_event_types, number_of_states, 1)
        return cy.log_likelihood_of_events_partial(event_type, float(base_rate[0]), impact_coefficients[:, :, 0],
                                                   decay_coefficients[:, :, 0],
                                                   number_of_event_types, number_of_states, times, events, states,
                                                   float(time_start), float(time_end))

    def gradient_partial(self, event_type, parameters, times, events, states, time_start, time_end):
        r"""
        Computes the gradient of the partial log-likelihood :math:`l_e` with respect to the
        parameters :math:`(\nu, \alpha, \beta)` that govern :math:`\lambda_e`, the intensity of events of type `e`.
        The method wraps a C implementation that was obtained via Cython.

        :type event_type: int
        :param event_type: `e`, the event type for which we want to compute the gradient of the partial
                            log-likelihood :math:`l_e`.
        :type parameters: 1D numpy array
        :param parameters: only the parameters :math:`(\nu, \alpha, \beta)` that govern :math:`\lambda_e`  put into a single
                           array.
                           Go from `parameters` to :math:`(\nu, \alpha, \beta)` and vice versa using
                           :py:meth:`~mpoints.hybrid_hawkes_exp.HybridHawkesExp.array_to_parameters`
                           and :py:meth:`~mpoints.hybrid_hawkes_exp.HybridHawkesExp.parameters_to_array`.
        :type times: 1D numpy array of float
        :param times: the times at which events occur.
        :type events: 1D numpy array of int
        :param events: the sequence of event types, `events[n]` is the event type of the `n` th event.
        :type states: 1D numpy array of int
        :param states: the sequence of states, `states[n]` is the new state of the system following the `n` th event.
        :type time_start: float
        :param time_start: :math:`t_0`, the time at which we consider that the process started, prior times are treated as an
                           initial condition.
        :type time_end: float
        :param time_end: :math:`T`, the time at which we stopped to record the process.
        :rtype: float
        :return: the gradient of the partial log-likelihood :math:`l_e`.
        """
        number_of_event_types = self.number_of_event_types
        number_of_states = self.number_of_states
        base_rate, impact_coefficients, decay_coefficients = \
            HybridHawkesExp.array_to_parameters(
                parameters, number_of_event_types, number_of_states, 1)
        g_base_rate, g_impact_coefficients, g_decay_coefficients = \
            cy.gradient_partial(event_type, float(base_rate[0]), impact_coefficients[:, :, 0], decay_coefficients[:, :, 0],
                                number_of_event_types,
                                number_of_states, times, events, states, float(time_start), float(time_end))
        a = np.zeros((number_of_event_types, number_of_states, 1))
        b = np.zeros((number_of_event_types, number_of_states, 1))
        a[:, :, 0] = g_impact_coefficients
        b[:, :, 0] = g_decay_coefficients
        return self.parameters_to_array([g_base_rate], a, b)

    'Miscellaneous tools'

    def intensities_of_events_at_times(self, compute_times, times, events, states):
        """
        Computes the intensities at the `compute_times` given a realisation of the state-dependent Hawkes process.

        :type compute_times: 1D numpy array of float
        :param compute_times: the times at which the intensities will be computed.
        :type times: 1D numpy array of float
        :param times: the times at which events occur.
        :type events: 1D numpy array of int
        :param events: the sequence of event types, `events[n]` is the event type of the `n` th event.
        :type states: 1D numpy array of int
        :param states: the sequence of states, `states[n]` is the new state of the system following the `n` th event.
        :rtype: 1D numpy array of float, 2D numpy array of float
        :return: the first array contains both the `compute_times` and the event times of the given realisation,
                 in increasing order. Note that the event times are put there twice:
                 the intensities are computed just before and right after the event times.
                 The second array gives the intensities at the times of the first array.
                 `array2[e,n]` is the intensity of events of type `e` at time `array1[n]`.
        """
        'Start/end time and first index of event times'
        # time at which we start to compute the intensity
        time_start = compute_times[0]
        time_end = compute_times[-1]
        next_event_time_index = bisect.bisect_right(
            times, time_start)  # first event time occurring in between
        next_event_time = time_end + 1  # in case there is no event after 'time_start'
        # i.e. there is an event time after
        if next_event_time_index < len(times):
            next_event_time = times[next_event_time_index]
        '''Initialise the partial sums S_{e',x',e} that will allow use to compute the intensity recursively'''
        partial_sums = np.zeros(
            (self.number_of_event_types, self.number_of_states, self.number_of_event_types))
        for n in range(next_event_time_index):
            time = times[n]
            event = events[n]
            state = states[n]
            for e in range(self.number_of_event_types):
                alpha = self.impact_coefficients[event, state, e]
                beta = self.decay_coefficients[event, state, e]
                partial_sums[event, state, e] += alpha * \
                    math.exp(-beta * (time_start - time))
        'Create aggregated sequence of times, events, and states (event times + compute times)'
        times_aggregated = []
        events_aggregated = []
        states_aggregated = []
        for t in compute_times:
            if t < next_event_time:
                times_aggregated.append(t)
                # -1 means that it is not an event time
                events_aggregated.append(-1)
                states_aggregated.append(-1)
            elif t > next_event_time:
                while next_event_time < t:
                    'We add it twice so that the intensity is computed just before and right after'
                    'This is to get nice looking figures when plotting the intensities'
                    times_aggregated.append(next_event_time)
                    times_aggregated.append(next_event_time)
                    event = events[next_event_time_index]
                    state = states[next_event_time_index]
                    events_aggregated.append(-1)
                    events_aggregated.append(event)
                    states_aggregated.append(-1)
                    states_aggregated.append(state)
                    next_event_time_index += 1
                    next_event_time = time_end + 1  # in case there is no event after 'time_start'
                    # i.e. there is an event time after
                    if next_event_time_index < len(times):
                        next_event_time = times[next_event_time_index]
                'Now add the time t'
                times_aggregated.append(t)
                events_aggregated.append(-1)
                states_aggregated.append(-1)
        'Compute the intensities at the aggregated times'
        number_of_times = len(times_aggregated)
        result_intensities = np.zeros(
            (self.number_of_event_types, number_of_times))
        intensities = self.intensities_of_events(partial_sums)
        for e in range(self.number_of_event_types):
            result_intensities[e, 0] = intensities[e]
        for n in range(1, number_of_times):
            time_increment = times_aggregated[n] - times_aggregated[n-1]
            event = events_aggregated[n]
            'Update partial sums: time decay effect'
            if time_increment > 0:
                for e1 in range(self.number_of_event_types):
                    for x in range(self.number_of_states):
                        for e2 in range(self.number_of_event_types):
                            beta = self.decay_coefficients[e1, x, e2]
                            partial_sums[e1, x,
                                         e2] *= math.exp(-beta * time_increment)
            'Update partial sums: impact of new event'
            if event >= 0:
                state = states_aggregated[n]
                for e2 in range(self.number_of_event_types):
                    alpha = self.impact_coefficients[event, state, e2]
                    partial_sums[event, state, e2] += alpha
            'Compute intensities and save'
            intensities = self.intensities_of_events(partial_sums)
            for e1 in range(self.number_of_event_types):
                result_intensities[e1, n] = intensities[e1]
        return times_aggregated, result_intensities

    def compute_partial_sums(self, times, events, states, time_end,
                             initial_partial_sums=None, time_initial_condition=None):
        r"""
        Computes the partial sums

        .. math::

            S_{e'xe}(-\infty,T] := \alpha_{e'xe} \sum_{n : t^{e'x}_n \leq T} \exp(-\beta_{e'xe}(T-t^{e'x}_n)),

        where `e` and `e'` are event types, `x` is a possible state, :math:`t^{e'x}_n` is the time when the `n` th
        event of type `e'` after which the state is `x` occurred.
        The collection of partial sums :math:`(S_{e'xe})` at time :math:`T` encodes the history of the
        state-dependent Hawkes process up to time :math:`T`. It can for example be used to simulate the process from
        time :math:`T`. In :py:meth:`~mpoints.hybrid_hawkes_exp.HybridHawkesExp.simulate`, one can pass the partial
        sums as an initial condition instead of the entire history.

        :type times: 1D numpy array of float
        :param times: the times at which events occur.
        :type events: 1D numpy array of int
        :param events: the sequence of event types, `events[n]` is the event type of the `n` th event.
        :type states: 1D numpy array of int
        :param states: the sequence of states, `states[n]` is the new state of the system following the `n` th event.
        :type time_end: float
        :param time_end: :math:`T`, the time up to which the partial sums are computed.
        :type initial_partial_sums: 3D numpy array of float
        :param initial_partial_sums: an initial condition (events that occur before `times`) can be given implicitly
                                     as partial sums.
        :type time_initial_condition: float
        :param time_initial_condition: the time up to which the partial sums given as an initial condition were
                                       computed.
        :rtype: 3D numpy array of float
        :return: the partial sums, `array[e', x, e]` corresponds to :math:`S_{e'x'e}`.
        """
        'Compute contribution of the given events'
        partial_sums = np.zeros(
            (self.number_of_event_types, self.number_of_states, self.number_of_event_types))
        index_end = bisect.bisect_right(times, time_end)
        for n in range(index_end):
            time = times[n]
            event = events[n]
            state = states[n]
            partial_sums[event, state,
                         :] += np.exp(- self.decay_coefficients[event, state, :] * (time_end - time))
        partial_sums = np.multiply(partial_sums, self.impact_coefficients)
        'Add contribution of the given initial condition'
        if np.shape(initial_partial_sums) != () and time_initial_condition != None:
            partial_sums += np.multiply(np.exp(- self.decay_coefficients * (time_end - time_initial_condition)),
                                        initial_partial_sums)
        return partial_sums

    def intensity_of_event(self, event_type, partial_sums):
        r"""
        Computes the intensity of events of type `event_type`, given the partial sums :math:`S_{e'xe}`.

        :type event_type: int
        :param event_type: the event type for which the intensity is computed.
        :type partial_sums: 3D numpy array of float
        :param partial_sums: the partial sums :math:`S_{e',x',e}` at the considered time.
        :rtype: float
        :return: the value of :math:`\lambda_e` at the considered time.
        """
        result = self.base_rates[event_type]
        for e in range(self.number_of_event_types):
            for x in range(self.number_of_states):
                result += partial_sums[e, x, event_type]
        return result

    def intensities_of_events(self, partial_sums):
        r"""
        Computes the intensities, given the partial sums :math:`S_{e'xe}`.

        :type partial_sums: 3D numpy array of float
        :param partial_sums: the partial sums :math:`S_{e',x',e}` at the considered time.
        :rtype: 1D numpy array of float
        :return: the intensities, `array[e]` is the value of :math:`\lambda_e` at the considered time.
        """
        result = np.zeros(self.number_of_event_types)
        for e in range(self.number_of_event_types):
            result[e] = self.intensity_of_event(e, partial_sums)
        return result

    @staticmethod
    def parameters_to_array(base_rates, impact_coefficients, decay_coefficients):
        r"""
        Puts the model parameters :math:`(\nu, \alpha, \beta)` into a one dimensional array.

        :type base_rates: 1D numpy array
        :param base_rates: the collection :math:`(\nu_{e})`.
        :type impact_coefficients: 3D numpy array
        :param impact_coefficients: the collection :math:`(\alpha_{e'xe})`.
        :type decay_coefficients: 3D numpy array
        :param decay_coefficients: the collection :math:`(\beta_{e'xe})`.
        :rtype: 1D numpy array
        :return: the parameters put into a single 1D array.
        """
        s = np.shape(impact_coefficients)
        number_of_event_types_1 = s[0]
        number_of_states = s[1]
        number_of_event_types_2 = s[2]
        result =\
            np.zeros(number_of_event_types_2 + 2 * number_of_event_types_1 *
                     number_of_event_types_2 * number_of_states)
        for n in range(number_of_event_types_2):
            result[n] = base_rates[n]
        for i in range(number_of_event_types_1):
            for j in range(number_of_states):
                for k in range(number_of_event_types_2):
                    index = number_of_event_types_2 + j * \
                        number_of_event_types_1 * number_of_event_types_2
                    index += i * number_of_event_types_2 + k
                    result[index] = impact_coefficients[i, j, k]
        for i in range(number_of_event_types_1):
            for j in range(number_of_states):
                for k in range(number_of_event_types_2):
                    index = number_of_event_types_2
                    index += number_of_event_types_1 * number_of_event_types_2 * number_of_states
                    index += j * number_of_event_types_1 * number_of_event_types_2
                    index += i * number_of_event_types_2 + k
                    result[index] = decay_coefficients[i, j, k]
        return result

    @staticmethod
    def array_to_parameters(array, number_of_event_types_1, number_of_states, number_of_event_types_2=0):
        r"""
        Retrieves the parameters :math:`(\nu, \alpha, \beta)` from a 1D array.
        It is NOT assumed that the length of the 1st and 3rd dimensions of the arrays :math:`(\alpha_{e'xe})` and
        :math:`(\beta_{e'xe})` are equal. For instance, in
        :py:meth:`~mpoints.hybrid_hawkes_exp.HybridHawkesExp.log_likelihood_of_events_partial`,
        only of subgroup of the parameters :math:`(\nu, \alpha, \beta)` are required.

        :type array: 1D numpy array
        :param array: an array containing the parameters :math:`(\nu, \alpha, \beta)`.
        :type number_of_event_types_1: int
        :param number_of_event_types_1: length of the first dimension of :math:`(\alpha_{e'xe})` and
                                        :math:`(\beta_{e'xe})`.
        :type number_of_states: int
        :param number_of_states: number of possible states in the state-dependent Hawkes process,
                                 length of the second dimension.
        :type number_of_event_types_2: int
        :param number_of_event_types_2: length of the third dimension of :math:`(\alpha_{e'xe})` and
                                        :math:`(\beta_{e'xe})`. It is also the length of :math:`\nu`.
                                        If set to zero, we assume that `array` contains ALL the parameters
                                        and not only a subgroup, meaning that
                                        1st and 3rd dimensions of the arrays :math:`(\alpha_{e'xe})` and
                                        :math:`(\beta_{e'xe})` are equal.
        :rtype: 1D numpy array, 3D numpy array, 3D numpy array
        :return: the parameters :math:`(\nu, \alpha, \beta)`.
        """
        if number_of_event_types_2 == 0:
            number_of_event_types_2 = number_of_event_types_1
        base_rates = np.zeros(number_of_event_types_2)
        for n in range(number_of_event_types_2):
            base_rates[n] = array[n]
        impact_coefficients = np.zeros(
            (number_of_event_types_1, number_of_states, number_of_event_types_2))
        for i in range(number_of_event_types_1):
            for j in range(number_of_states):
                for k in range(number_of_event_types_2):
                    index = number_of_event_types_2 + j * \
                        number_of_event_types_1 * number_of_event_types_2
                    index += i * number_of_event_types_2 + k
                    impact_coefficients[i, j, k] = array[index]
        decay_coefficients = np.zeros(
            (number_of_event_types_1, number_of_states, number_of_event_types_2))
        for i in range(number_of_event_types_1):
            for j in range(number_of_states):
                for k in range(number_of_event_types_2):
                    index = number_of_event_types_2
                    index += number_of_event_types_1 * number_of_event_types_2 * number_of_states
                    index += j * number_of_event_types_1 * number_of_event_types_2
                    index += i * number_of_event_types_2 + k
                    decay_coefficients[i, j, k] = array[index]
        return base_rates, impact_coefficients, decay_coefficients

    @staticmethod
    def transition_matrix_to_string(transition_probabilities):
        r"""
        Transforms the transition probabilities :math:`\phi` to a string that can be printed in the console or a
        text file.

        :type transition_probabilities: 3D numpy array
        :param transition_probabilities: the transition probabilities :math:`\phi`.
        :rtype: string
        :return: sequence of transition probability matrices :math:`(\phi_e)`, one per event type `e`.
        """
        number_of_states = len(transition_probabilities)
        number_of_events = len(transition_probabilities[0])
        result = ''
        for e in range(number_of_events):
            m = np.zeros((number_of_states, number_of_states))
            for x1 in range(number_of_states):
                for x2 in range(number_of_states):
                    m[x1][x2] = transition_probabilities[x1][e][x2]
            result += str(m)
            if e < number_of_events - 1:
                result += '\n'
        return result

    @staticmethod
    def impact_coefficients_to_string(impact_coefficients):
        r"""
        Transforms the impact coefficients :math:`\alpha` to a string that can be printed in the console or a text file.

        :type impact_coefficients: 3D numpy array
        :param impact_coefficients: the impact coefficients :math:`\alpha`.
        :rtype: string
        :return: sequence of matrices :math:`(\alpha_{\cdot x \cdot})`, one per possible state `x`.
        """
        return HybridHawkesExp.transition_matrix_to_string(impact_coefficients)

    @staticmethod
    def decay_coefficients_to_string(decay_coefficients):
        r"""
        Transforms the decay coefficients :math:`\beta` to a string that can be printed in the console or a text file.

        :type decay_coefficients: 3D numpy array
        :param decay_coefficients: the decay coefficients :math:`\beta`.
        :rtype: string
        :return: sequence of matrices :math:`(\beta_{\cdot x \cdot})`, one per possible state `x`.
        """
        return HybridHawkesExp.transition_matrix_to_string(decay_coefficients)

    @staticmethod
    def proportion_of_events_and_states(events, states, number_of_event_types, number_of_states):
        r"""
        Computes the empirical distribution of the events and states.

        :type events: 1D numpy array of int
        :param events: the sequence of event types, `events[n]` is the event type of the `n` th event.
        :type states: 1D numpy array of int
        :param states: the sequence of states, `states[n]` is the new state of the system following the `n` th event.
        :type number_of_event_types: int
        :param number_of_event_types: number of different event types.
        :type number_of_states: int
        :param number_of_states: number of possible states.
        :rtype: 2D numpy array of float
        :return: `array[e,x]` is the percentage of events of type `e` after which the state is `x`.
        """
        proportion_events_states = np.zeros(
            (number_of_event_types, number_of_states))
        size = len(events)
        for n in range(size):
            e = events[n]
            x = states[n]
            proportion_events_states[e, x] += 1
        proportion_events_states = np.divide(proportion_events_states, size)
        return proportion_events_states

    def generate_base_rates_labels(self):
        r"""
        Produces labels for the base rates :math:`\nu`.
        This uses the events labels of the model.

        :rtype: list of string
        :return: `list[e]` returns a label for :math:`\nu_e`.
        """
        labels = []
        for e in range(self.number_of_event_types):
            l = r'$\nu_{' + self.events_labels[e] + '}$'
            labels.append(l)
        return labels

    def generate_impact_coefficients_labels(self):
        r"""
        Produces labels for the impact coefficients :math:`\alpha`.
        This uses the events and states labels of the model.

        :rtype: list of list of list of string
        :return: `list[e'][x][e]` returns a label for :math:`\alpha_{e'xe}`.
        """
        labels = []
        for e1 in range(self.number_of_event_types):
            l1 = []
            for x in range(self.number_of_states):
                l2 = []
                for e2 in range(self.number_of_event_types):
                    s = r'$\alpha_{' + self.events_labels[e1]
                    s += r' \rightarrow ' + self.events_labels[e2] + '}('
                    s += self.states_labels[x] + ')$'
                    l2.append(s)
                l1.append(l2)
            labels.append(l1)
        return labels

    def generate_decay_coefficients_labels(self):
        r"""
        Produces labels for the decay coefficients :math:`\beta`.
        This uses the events and states labels of the model.

        :rtype: list of list of list of string
        :return: `list[e'][x][e]` returns a label for :math:`\beta_{e'xe}`.
        """
        labels = []
        for e1 in range(self.number_of_event_types):
            l1 = []
            for x in range(self.number_of_states):
                l2 = []
                for e2 in range(self.number_of_event_types):
                    s = r'$\beta_{' + self.events_labels[e1]
                    s += r' \rightarrow ' + self.events_labels[e2] + '}('
                    s += self.states_labels[x] + ')$'
                    l2.append(s)
                l1.append(l2)
            labels.append(l1)
        return labels

    def generate_product_labels(self):
        r"""
        Produces labels for all the couples of events types and possible states.
        This uses the events and states labels of the model.

        :rtype: list of strings
        :return: the label for the couple `(e, x)` is given by by `list[n]` where
                 :math:`n = x + e \times \mbox{number_of_event_types}`.
        """
        r = []
        for e in self.events_labels:
            for x in self.states_labels:
                r.append(e + ', ' + x)
        return r



import napiod
import numpy as np
import matplotlib.pyplot as plt


def main():

    # Randomly generate the parameters for the order book
    phis = napiod.model.generate_random_transition_probabilities(
        with_agent=False)
    nus, alphas, betas = napiod.model.generate_random_hawkes_parameters(
        with_agent=False)

    # Define the parameters of the agent
    agent_base_rate = .25
    agent_self_excitation = np.zeros(
        (napiod.model.TOTAL_NUMBER_OF_EVENT_TYPES,
         napiod.model.TOTAL_NUMBER_OF_STATES),
        dtype=float)
    agent_self_excitation[
        :,
        list(
            napiod.model.state_indexes_with_negative_imbalances())] = 1.
    agent_self_decay = 2.0 * np.ones(
        (napiod.model.TOTAL_NUMBER_OF_EVENT_TYPES,
         napiod.model.TOTAL_NUMBER_OF_STATES),
        dtype=float)
    agent_impact_on_others = np.zeros(
        (napiod.model.TOTAL_NUMBER_OF_STATES,
         napiod.model.NUMBER_OF_ORDERBOOK_EVENTS),
        dtype=float
    )
    agent_impact_on_others[:, 0] = .1
    agent_impact_on_others[:, 1] = .5
    agent_impact_on_others[:, 3] = .15
    agent_decay_on_others = 0.85 * np.ones(
        (napiod.model.TOTAL_NUMBER_OF_STATES,
         napiod.model.NUMBER_OF_ORDERBOOK_EVENTS),
        dtype=float
    )
    agent_state_transition = np.zeros(
        (napiod.model.TOTAL_NUMBER_OF_STATES,
         napiod.model.TOTAL_NUMBER_OF_STATES),
        dtype=float
    )
    agent_state_transition[:, [0, 1, 2,  3]] = .25

    # Initialise the price impact model
    direction = -1
    price_impact = napiod.model.PriceImpact(direction)
    # Set the parameters of the order book and of the agent
    price_impact.set_orderbook_and_agent(
        phis,
        nus,
        alphas,
        betas,
        agent_state_transition,
        agent_base_rate,
        agent_impact_on_others,
        agent_self_excitation,
        agent_decay_on_others,
        agent_self_decay,
    )

    # Simulate the dynamic evolution of the order book while the agent executes their order
    execution_start = 0.
    execution_end = 100.
    horizon = 200.
    initial_state = 3
    times, events, states = price_impact.simulate_execution(
        execution_start,
        execution_end,
        horizon,
        initial_state=initial_state
    )

    # Reconstruct the price path from the simulated states of the model
    price = napiod.model.price_path_from_states(0, states)  # NOQA

    # Compute the agent price impact profile
    # Note: this assumes execution_start = 0.
    impact_profile = price_impact.compute_price_impact_profile(
        times, events, states,
        horizon,
        execution_end,
        dt=.5,
    )

    # Visualisations
    plot_time_start = 95.
    plot_time_end = 105.

    # Plot the price path from the simulated states of the model
    fig_price, _ = napiod.plot.price_path(
        times,
        events,
        states,
        plot_time_start,
        plot_time_end,
        initial_price=0
    )

    # Plot the evolution of the order book as encoded in the state-dependent Hawkes process
    fig_sample, _ = napiod.plot.hawkes_sample(
        times,
        events,
        states,
        price_impact.orderbook,
        plot_time_start,
        plot_time_end,
    )

    # Plot the agent's price impact profile
    fig_impact, _ = napiod.plot.price_impact_profile(
        impact_profile,
        plot_time_start,
        plot_time_end
    )

    return fig_price, fig_sample, fig_impact


if __name__ == '__main__':
    fig_price, fig_sample, fig_impact = main()
    plt.show()



import setuptools
import numpy

ext_modules = [
    setuptools.Extension('napiod.intensities',
                         sources=['napiod/intensities.c'],
                         extra_compile_args=["-ffast-math"],
                         ),
    setuptools.Extension('napiod.impact',
                         sources=['napiod/impact.c'],
                         extra_compile_args=["-ffast-math"],
                         ),
    setuptools.Extension('napiod.model',
                         sources=['napiod/model.c'],
                         extra_compile_args=["-ffast-math"],
                         )
]

setuptools.setup(
    name="napiod",
    description="Non-average price impact in order-driven markets",
    ext_modules=ext_modules,
    include_dirs=[numpy.get_include()],
)
